from copy import copy
from typing import Dict, List, Union

import numpy as np
import pandas as pd
from sklearn.preprocessing import LabelEncoder

from fedot.core.data.data import InputData, OutputData, data_type_is_table, data_type_is_ts, data_type_is_text
from fedot.core.data.data_preprocessing import (
    data_has_categorical_features,
    data_has_missing_values,
    find_categorical_columns,
    replace_inf_with_nans,
    replace_nans_with_empty_strings
)
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.log import default_log
from fedot.core.operations.evaluation.operation_implementations.data_operations.categorical_encoders import (
    LabelEncodingImplementation,
    OneHotEncodingImplementation
)
from fedot.core.operations.evaluation.operation_implementations.data_operations.sklearn_transformations import (
    ImputationImplementation
)
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.preprocessing.categorical import BinaryCategoricalPreprocessor
from fedot.preprocessing.data_types import NAME_CLASS_INT, TableTypesCorrector
from fedot.preprocessing.data_type_check import exclude_ts, exclude_multi_ts, exclude_image
from fedot.preprocessing.structure import DEFAULT_SOURCE_NAME, PipelineStructureExplorer
# The allowed percent of empty samples in features.
# Example: 90% objects in features are 'nan', then drop this feature from data.
ALLOWED_NAN_PERCENT = 0.9


class DataPreprocessor:
    """
    Class which contains methods for data preprocessing.
    The class performs two types of preprocessing: obligatory and optional

    obligatory - delete rows where nans in the target, remove features,
    which full of nans, delete extra_spaces
    optional - depends on what operations are in the pipeline, gap-filling
    is applied if there is no imputation operation in the pipeline, categorical
    encoding is applied if there is no encoder in the structure of the pipeline etc.
    """

    def __init__(self):
        # There was performed encoding for string target column or not
        self.target_encoders: Dict[str, LabelEncoder] = {}
        self.features_encoders: Dict[str, Union[OneHotEncodingImplementation, LabelEncodingImplementation]] = {}
        self.use_label_encoder: bool = False
        self.features_imputers: Dict[str, ImputationImplementation] = {}
        self.ids_relevant_features: Dict[str, List[int]] = {}

        # Cannot be processed due to incorrect types or large number of nans
        self.ids_incorrect_features: Dict[str, List[int]] = {}
        # Categorical preprocessor for binary categorical features
        self.binary_categorical_processors: Dict[str, BinaryCategoricalPreprocessor] = {}
        self.types_correctors: Dict[str, TableTypesCorrector] = {}
        self.main_target_source_name = None

        self.log = default_log(self)

    def __setstate__(self, state):
        # Implemented for backward compatibility for unpickling
        #  Pipelines with older preprocessor that had DiGraph with Nodes inside.
        #  see https://github.com/nccr-itmo/FEDOT/pull/802
        unrelevant_fields = ['structure_analysis']
        for field in unrelevant_fields:
            if field in state:
                del state[field]
        self.__dict__.update(state)

    def _init_supplementary_preprocessors(self, data: Union[InputData, MultiModalData]):
        """ Initialize helpers for preprocessor

        :param data: data class with input data for preprocessing
        """
        categorical_sources = list(self.binary_categorical_processors)
        types_sources = list(self.types_correctors)
        if len(categorical_sources) == len(types_sources) and types_sources:
            # Preprocessors have been already initialized
            return None
        self.helpers_were_initialized = True

        self.binary_categorical_processors = {}
        self.types_correctors = {}

        if isinstance(data, InputData):
            self.binary_categorical_processors[DEFAULT_SOURCE_NAME] = BinaryCategoricalPreprocessor()
            self.types_correctors[DEFAULT_SOURCE_NAME] = TableTypesCorrector()
        elif isinstance(data, MultiModalData):
            for data_source in data:
                self.binary_categorical_processors[data_source] = BinaryCategoricalPreprocessor()
                self.types_correctors[data_source] = TableTypesCorrector()
        else:
            raise ValueError('Unknown type of data.')

    def _init_main_target_source_name(self, multi_data: MultiModalData):
        """ Define for MultiModal data branches with main target and side ones """
        if self.main_target_source_name is not None:
            # Target name has been already defined
            return None

        for data_source_name, input_data in multi_data.items():
            if input_data.supplementary_data.is_main_target:
                self.main_target_source_name = data_source_name
                break

    def obligatory_prepare_for_fit(self, data: Union[InputData, MultiModalData]):
        """
        Perform obligatory preprocessing for pipeline fit method.
        It includes removing features full of nans, extra spaces in features deleting,
        drop rows where target cells are none
        """
        # TODO add advanced gapfilling for time series and advanced gap-filling
        self._init_supplementary_preprocessors(data)

        if isinstance(data, InputData):
            data = self._prepare_obligatory_unimodal_for_fit(data, source_name=DEFAULT_SOURCE_NAME)

        elif isinstance(data, MultiModalData):
            self._init_main_target_source_name(data)
            for data_source_name, values in data.items():
                data[data_source_name] = self._prepare_obligatory_unimodal_for_fit(values,
                                                                                   source_name=data_source_name)

        self.mark_as_preprocessed(data)
        return data

    def obligatory_prepare_for_predict(self, data: Union[InputData, MultiModalData]):
        """ Perform obligatory preprocessing for pipeline predict method """
        if isinstance(data, InputData):
            data = self._prepare_obligatory_unimodal_for_predict(data, source_name=DEFAULT_SOURCE_NAME)

        elif isinstance(data, MultiModalData):
            for data_source_name, values in data.items():
                data[data_source_name] = self._prepare_obligatory_unimodal_for_predict(values,
                                                                                       source_name=data_source_name)

        self.mark_as_preprocessed(data)
        return data

    def optional_prepare_for_fit(self, pipeline, data: Union[InputData, MultiModalData]):
        """ Launch preprocessing operations if it is necessary for pipeline fitting

        :param pipeline: pipeline to prepare data for
        :param data: data to preprocess
        """
        self._init_supplementary_preprocessors(data)

        if isinstance(data, InputData):
            self._prepare_optional(pipeline, data, DEFAULT_SOURCE_NAME)
        else:
            # Multimodal data
            self._init_main_target_source_name(data)
            for data_source_name, values in data.items():
                self._prepare_optional(pipeline, values, data_source_name)

        return data

    def optional_prepare_for_predict(self, pipeline, data: Union[InputData, MultiModalData]):
        """ Launch preprocessing operations if it is necessary for pipeline predict stage.
        Preprocessor should already must be fitted.

        :param pipeline: pipeline to prepare data for
        :param data: data to preprocess
        """
        if isinstance(data, InputData):
            self._prepare_optional(pipeline, data, DEFAULT_SOURCE_NAME)
        else:
            # Multimodal data
            for data_source_name, values in data.items():
                self._prepare_optional(pipeline, values, data_source_name)

        return data

    def take_only_correct_features(self, data: InputData, source_name: str):
        """ Take only correct features in the table """
        current_relevant_ids = self.ids_relevant_features[source_name]
        if current_relevant_ids:
            data.features = data.features[:, current_relevant_ids]

    @exclude_ts
    @exclude_multi_ts
    @exclude_image
    def _prepare_obligatory_unimodal_for_fit(self, data: InputData, source_name: str) -> InputData:
        """ Method process InputData for pipeline fit method """
        if data.supplementary_data.was_preprocessed:
            # Preprocessing was already done - return data
            return data

        # Wrap indices in numpy array
        data.idx = np.array(data.idx)

        # Fix tables / time series sizes
        data = self._correct_shapes(data)
        replace_inf_with_nans(data)

        # Find incorrect features which must be removed
        self._find_features_full_of_nans(data, source_name)
        self.take_only_correct_features(data, source_name)
        data = self._drop_rows_with_nan_in_target(data)

        # Column types processing - launch after correct features selection
        self.types_correctors[source_name].convert_data_for_fit(data)
        if self.types_correctors[source_name].target_converting_has_errors:
            data = self._drop_rows_with_nan_in_target(data)

        # Train Label Encoder for categorical target if necessary and apply it
        self._train_target_encoder(data, source_name)
        data.target = self._apply_target_encoding(data, source_name)

        # TODO andreygetmanov target encoding must be obligatory for all data types
        if data_type_is_text(data):
            # TODO andreygetmanov to new class text preprocessing?
            replace_nans_with_empty_strings(data)
        elif data_type_is_table(data):
            data = self._clean_extra_spaces(data)
            # Process binary categorical features
            self.binary_categorical_processors[source_name].fit(data)
            data = self.binary_categorical_processors[source_name].transform(data)

        return data

    @exclude_ts
    @exclude_multi_ts
    @exclude_image
    def _prepare_obligatory_unimodal_for_predict(self, data: InputData, source_name: str) -> InputData:
        """ Method process InputData for pipeline predict method """
        if data.supplementary_data.was_preprocessed:
            # Preprocessing was already done - return data
            return data

        # Wrap indices in numpy array
        data.idx = np.array(data.idx)

        # Fix tables / time series sizes
        data = self._correct_shapes(data)
        replace_inf_with_nans(data)

        # Perform preprocessing for types - launch after correct features selection
        self.take_only_correct_features(data, source_name)
        self.types_correctors[source_name].convert_data_for_predict(data)

        if data_type_is_text(data):
            replace_nans_with_empty_strings(data)
        if data_type_is_table(data):
            data = self._clean_extra_spaces(data)
            data = self.binary_categorical_processors[source_name].transform(data)

        return data

    def _prepare_optional(self, pipeline, data: InputData, source_name: str):
        """ Perform optional fitting/preprocessing for unimodal data """
        if not data_type_is_table(data):
            return data

        if data_has_missing_values(data):
            # Data contains missing values
            has_imputer = PipelineStructureExplorer.\
                check_structure_by_tag(pipeline, tag_to_check='imputation', source_name=source_name)
            if not has_imputer:
                data = self._apply_imputation_unidata(data, source_name)

        if data_has_categorical_features(data):
            # Data contains categorical features values
            has_encoder = PipelineStructureExplorer.\
                check_structure_by_tag(pipeline, tag_to_check='encoding', source_name=source_name)
            if not has_encoder:
                data = self._apply_categorical_encoding(data, source_name)

    def _find_features_full_of_nans(self, data: InputData, source_name: str):
        """ Find features with more than ALLOWED_NAN_PERCENT nan's

        :param data: data to find columns with nan values
        :param source_name: name of data source node
        """
        # Initialize empty lists to fill it with indices
        self.ids_relevant_features[source_name] = []
        self.ids_incorrect_features[source_name] = []

        features = data.features
        n_samples, n_columns = features.shape

        for i in range(n_columns):
            feature = features[:, i]
            if np.sum(pd.isna(feature)) / n_samples < ALLOWED_NAN_PERCENT:
                self.ids_relevant_features[source_name].append(i)
            else:
                self.ids_incorrect_features[source_name].append(i)

    @staticmethod
    def _drop_rows_with_nan_in_target(data: InputData):
        """ Drop rows where in target column there are nans """
        features = data.features
        target = data.target

        # Find indices of nans rows. Using pd instead of np because it is needed for string columns
        bool_target = np.array(pd.isna(target))
        number_nans_per_rows = bool_target.sum(axis=1)

        # Ids of rows which doesn't contain nans in target
        non_nan_row_ids = np.ravel(np.argwhere(number_nans_per_rows == 0))

        if len(non_nan_row_ids) == 0:
            raise ValueError('Data contains too much nans in the target column(s)')
        data.features = features[non_nan_row_ids, :]
        data.target = target[non_nan_row_ids, :]
        data.idx = np.array(data.idx)[non_nan_row_ids]

        return data

    @staticmethod
    def _clean_extra_spaces(data: InputData):
        """ Remove extra spaces from data.
            Transform cells in columns from ' x ' to 'x'
        """
        features = pd.DataFrame(data.features)
        features = features.applymap(lambda x: x.strip() if isinstance(x, str) else x)

        data.features = np.array(features)
        return data

    def label_encoding_for_fit(self, data: InputData, source_name: str = DEFAULT_SOURCE_NAME):
        """
        Encode categorical features to numerical using LabelEncoder. In additional,
        save encoders to use later for prediction data.

        :param data: data to transform
        :param source_name: name of data source node
        """
        if data_has_categorical_features(data):
            encoder = self.features_encoders.get(source_name)
            if not isinstance(encoder, LabelEncodingImplementation) or encoder is None:
                encoder = LabelEncodingImplementation()
                encoder.fit(data)
                # Store encoder to make prediction in the future
                self.features_encoders.update({source_name: encoder})
                self.use_label_encoder = True
            encoder_output = encoder.transform_for_fit(data)
            data.features = encoder_output.predict
            data.supplementary_data = encoder_output.supplementary_data

    def cut_dataset(self, data: InputData, border: int):
        """ Cutting large dataset based on border (number of objects to remain) """
        self.log.info("Cut dataset due to it size is large")
        # TODO: don't shuffle the data here, because it is done in GPComposer
        data.shuffle()
        data.idx = data.idx[:border]
        data.features = data.features[:border]
        data.target = data.target[:border]

    def _apply_imputation_unidata(self, data: InputData, source_name: str):
        """ Fill in the gaps in the data inplace.

        :param data: data for fill in the gaps
        """
        imputer = self.features_imputers.get(source_name)
        if not imputer:
            imputer = ImputationImplementation()
            output_data = imputer.fit_transform(data)
            self.features_imputers[source_name] = imputer
        else:
            output_data = imputer.transform(data)
        data.features = output_data.predict
        return data

    def _apply_categorical_encoding(self, data: InputData, source_name: str):
        """
        Transforms the data inplace. Uses the same transformations as for the training data if trained already.
        Otherwise fits appropriate encoder and converts data's categorical features with it.

        :param data: data to transformation
        :param source_name: name of data source node
        """
        encoder = self.features_encoders.get(source_name)
        if encoder is None:
            encoder = LabelEncodingImplementation() if self.use_label_encoder else OneHotEncodingImplementation()
            encoder.fit(data)
            self.features_encoders[source_name] = encoder
        output_data = encoder.transform_for_fit(data)
        output_data.predict = output_data.predict.astype(float)
        data.features = output_data.predict
        data.supplementary_data = output_data.supplementary_data
        return data

    def _train_target_encoder(self, data: InputData, source_name: str):
        """ Convert string categorical target into integer column using LabelEncoder """
        categorical_ids, _ = find_categorical_columns(data.target, data.supplementary_data.column_types['target'])

        if categorical_ids:
            # Target is categorical
            target_encoder = LabelEncoder()
            target_encoder.fit(data.target)
            self.target_encoders[source_name] = target_encoder

    def _apply_target_encoding(self, data: InputData, source_name: str) -> np.array:
        """ Apply trained encoder for target column

        For example, target [['red'], ['green'], ['red']] will be converted into
        [[0], [1], [0]]
        """
        if source_name in self.target_encoders:
            # Target encoders have already been fitted
            data.supplementary_data.column_types['target'] = [NAME_CLASS_INT]
            encoded_target = self.target_encoders[source_name].transform(data.target)
            if len(encoded_target.shape) == 1:
                encoded_target = encoded_target.reshape((-1, 1))
            return encoded_target
        else:
            return data.target

    def apply_inverse_target_encoding(self, column_to_transform: np.array) -> np.array:
        """ Apply inverse Label Encoding operation for target column """
        main_target_source_name = self._determine_target_converter()

        if main_target_source_name in self.target_encoders:
            # Check if column contains string objects
            categorical_ids, _ = find_categorical_columns(column_to_transform)
            if categorical_ids:
                # There is no need to perform converting (it was performed already)
                return column_to_transform
            # It is needed to apply fitted encoder to apply inverse transformation
            transformed = self.target_encoders[main_target_source_name].inverse_transform(column_to_transform)

            # Convert one-dimensional array into column
            if len(transformed.shape) == 1:
                transformed = transformed.reshape((-1, 1))
            return transformed
        else:
            # Return source column
            return column_to_transform

    def _determine_target_converter(self):
        """
        For inverse target transformation there is a need to determine
        which target encoder to use (if there are several targets in
        single MultiModal pipeline).
        """
        # Choose data source node name with main target
        if self.main_target_source_name is None:
            return DEFAULT_SOURCE_NAME
        else:
            return self.main_target_source_name

    @staticmethod
    def _correct_shapes(data: InputData) -> InputData:
        """
        Correct shapes of tabular data or time series: tabular must be
        two-dimensional arrays, text - array of (n, 1), time series - one-dim array
        """
        if data_type_is_table(data) or data.data_type is DataTypesEnum.multi_ts:
            if np.ndim(data.features) < 2:
                data.features = data.features.reshape((-1, 1))
            if data.target is not None and np.ndim(data.target) < 2:
                data.target = data.target.reshape((-1, 1))
        elif data_type_is_text(data):
            data.features = data.features.reshape((-1, 1))
            if data.target is not None and np.ndim(data.target) < 2:
                data.target = np.array(data.target).reshape((-1, 1))
        elif data_type_is_ts(data):
            data.features = np.ravel(data.features)

        return data

    @staticmethod
    def convert_indexes_for_fit(pipeline, data: Union[InputData, MultiModalData]):
        if isinstance(data, MultiModalData):
            for data_source_name in data:
                if data_type_is_ts(data[data_source_name]):
                    data[data_source_name] = data[data_source_name].convert_non_int_indexes_for_fit(pipeline)
            return data
        elif data_type_is_ts(data):
            return data.convert_non_int_indexes_for_fit(pipeline)
        else:
            return data

    @staticmethod
    def convert_indexes_for_predict(pipeline, data: Union[InputData, MultiModalData]):
        if isinstance(data, MultiModalData):
            for data_source_name in data:
                if data_type_is_ts(data[data_source_name]):
                    data[data_source_name] = data[data_source_name].convert_non_int_indexes_for_predict(pipeline)
            return data
        elif data_type_is_ts(data):
            return data.convert_non_int_indexes_for_predict(pipeline)
        else:
            return data

    @staticmethod
    def restore_index(input_data: InputData, result: OutputData):
        if isinstance(input_data, InputData):
            if input_data.supplementary_data.non_int_idx is not None:
                result.idx = copy(input_data.supplementary_data.non_int_idx)
                result.supplementary_data.non_int_idx = copy(input_data.idx)
        return result

    @staticmethod
    def mark_as_preprocessed(data: Union[InputData, MultiModalData]):
        if isinstance(data, InputData):
            data.supplementary_data.was_preprocessed = True
        else:
            # Multimodal data
            for values in data.values():
                values.supplementary_data.was_preprocessed = True


def merge_preprocessors(api_preprocessor: DataPreprocessor,
                        pipeline_preprocessor: DataPreprocessor) -> DataPreprocessor:
    """
    Combining two preprocessor objects. One is the preprocessor from the API,
    the second is the preprocessor from the obtained pipeline
    """
    # Take all obligatory data preprocessing from API
    new_data_preprocessor = api_preprocessor

    # Update optional preprocessing (take it from obtained pipeline)
    if not new_data_preprocessor.features_encoders:
        # Store features encoder from obtained pipeline because in API there are no encoding
        new_data_preprocessor.features_encoders = pipeline_preprocessor.features_encoders
    return new_data_preprocessor


def update_indices_for_time_series(test_data: Union[InputData, MultiModalData]):
    """ Replace indices for time series for predict stage """
    if test_data.task.task_type is not TaskTypesEnum.ts_forecasting:
        return test_data

    if isinstance(test_data, MultiModalData):
        # Process multimodal data - change indices in every data block
        for input_data in test_data.values():
            forecast_len = input_data.task.task_params.forecast_length
            if forecast_len < len(input_data.idx):
                # Indices incorrect - there is a need to reassign them
                last_id = len(input_data.idx)
                input_data.idx = np.arange(last_id, last_id + input_data.task.task_params.forecast_length)
    else:
        # Simple input data
        forecast_len = test_data.task.task_params.forecast_length
        if forecast_len < len(test_data.idx):
            last_id = len(test_data.idx)
            test_data.idx = np.arange(last_id, last_id + test_data.task.task_params.forecast_length)
    return test_data



from __future__ import annotations
from copy import copy

import numpy as np
import pandas as pd

from fedot.core.log import LoggerAdapter, default_log
from fedot.core.repository.tasks import Task, TaskTypesEnum


NoneType = type(None)
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from fedot.core.data.data import InputData

NAME_CLASS_STR = "<class 'str'>"
NAME_CLASS_INT = "<class 'int'>"
NAME_CLASS_FLOAT = "<class 'float'>"
NAME_CLASS_NONE = "<class 'NoneType'>"
FEDOT_STR_NAN = 'fedot_nan'
# If unique values in the feature column is less than 13 - convert column into string type else to numerical
CATEGORICAL_MAX_UNIQUE_TH = 13
# column must be removed if failed rate is between these constants below
# because it means that in the column there are approximately the same number of truly string and ints/floats
ACCEPTABLE_CONVERSION_FAILED_RATE_BOTTOM = 0.4
ACCEPTABLE_CONVERSION_FAILED_RATE_TOP = 0.65


class TableTypesCorrector:
    """
    Class for checking types in input data. Also perform conversion for columns with types conflicts
    """

    def __init__(self):
        # Threshold to convert numerical into categorical column
        self.categorical_max_uniques_th = CATEGORICAL_MAX_UNIQUE_TH

        self.acceptable_failed_rate_bottom = ACCEPTABLE_CONVERSION_FAILED_RATE_BOTTOM
        self.acceptable_failed_rate_top = ACCEPTABLE_CONVERSION_FAILED_RATE_TOP

        self.features_columns_info = {}
        self.target_columns_info = {}

        # Dictionary with information about converted during fitting columns
        self.features_converted_columns = {}
        self.target_converted_columns = {}

        # Columns to delete due to types conflicts
        self.columns_to_del = []
        # Column ids for transformation due to number of unique values
        self.numerical_into_str = []
        self.categorical_into_float = []

        # Indices of columns with filed string into numerical transformation
        self.string_columns_transformation_failed = {}

        # Is target column contains non-numerical cells during conversion
        self.target_converting_has_errors = False

        # Lists with column types for converting calculated on source input data
        self.features_types = None
        self.target_types = None
        self.log = default_log(self)

    def convert_data_for_fit(self, data: InputData):
        """ If column contain several data types - perform correction procedure """
        # Convert features to have an ability to insert str into float table or vice versa
        data.features = data.features.astype(object)

        # Determine types for each column in features and target if it is necessary
        self.features_columns_info = define_column_types(data.features)
        self.target_columns_info = define_column_types(data.target)

        # Correct types in features table
        data.features = self.features_types_converting(features=data.features)
        # Remain only correct columns
        data.features = self.remove_incorrect_features(data.features, self.features_converted_columns)

        # And in target(s)
        data.target = self.target_types_converting(target=data.target, task=data.task)
        data.supplementary_data.column_types = self.prepare_column_types_info(predictors=data.features,
                                                                              target=data.target,
                                                                              task=data.task)

        self._into_numeric_features_transformation_for_fit(data)
        # Launch conversion float and integer features into categorical
        self._into_categorical_features_transformation_for_fit(data)
        # Save info about features and target types
        self.features_types = copy(data.supplementary_data.column_types['features'])
        self.target_types = copy(data.supplementary_data.column_types['target'])

        self._retain_columns_info_without_types_conflicts(data)
        return data

    def convert_data_for_predict(self, data: InputData):
        """ Prepare data for predict stage. Include only column types transformation """
        # Ordering is important because after removing incorrect features - indices are obsolete
        data.features = data.features.astype(object)
        data.features = self.remove_incorrect_features(data.features, self.features_converted_columns)
        data.features = apply_type_transformation(data.features, self.features_types, self.log)
        data.target = apply_type_transformation(data.target, self.target_types, self.log)
        data.supplementary_data.column_types = self.prepare_column_types_info(predictors=data.features,
                                                                              target=data.target,
                                                                              task=data.task)

        # Convert column types
        self._into_numeric_features_transformation_for_predict(data)
        self._into_categorical_features_transformation_for_predict(data)
        self._retain_columns_info_without_types_conflicts(data)
        return data

    def remove_incorrect_features(self, table: np.array, converted_columns: dict):
        """
        Remove from the table columns with conflicts with types were not resolved

        :param table: tabular dataset based on which new dataset will be generated
        :param converted_columns: dictionary with actions with table
        """
        if not converted_columns:
            return table

        self.columns_to_del = [column_id for column_id, new_type_name in converted_columns.items() if
                               new_type_name == 'removed']
        if not self.columns_to_del:
            # There are no columns to delete
            return table

        # Remove all "bad" columns
        table = np.delete(table, self.columns_to_del, 1)
        return table

    def features_types_converting(self, features: np.array) -> np.array:
        """ Convert all elements in the data in every feature column into one type

        :param features: tabular features array
        """
        features_with_mixed_types = find_mixed_types_columns(self.features_columns_info)

        if not features_with_mixed_types:
            return features

        # There are mixed-types columns in features table - convert them
        for mixed_column_id in features_with_mixed_types:
            column_info = self.features_columns_info[mixed_column_id]

            if column_info.get('str_number') > 0 or column_info.get('float_number') > 0:
                # There are string elements in the array
                mixed_column = features[:, mixed_column_id]
                updated_column, new_type_name = self._convert_feature_into_one_type(mixed_column, column_info,
                                                                                    mixed_column_id)
                # Store information about converted columns
                self.features_converted_columns.update({mixed_column_id: new_type_name})

                if updated_column is not None:
                    features[:, mixed_column_id] = updated_column

        return features

    def target_types_converting(self, target: np.array, task: Task) -> np.array:
        """ Convert all elements in every target column into one type

        :param target: tabular target array
        :param task: task to solve
        """
        target_with_mixed_types = find_mixed_types_columns(self.target_columns_info)

        if not target_with_mixed_types:
            return target

        # There are mixed-types columns in features table - convert them
        for mixed_column_id in target_with_mixed_types:
            column_info = self.target_columns_info[mixed_column_id]

            if column_info.get('str_number') > 0:
                # There are string elements in the array
                mixed_column = target[:, mixed_column_id]
                updated_column, new_type_name = self._convert_target_into_one_type(mixed_column, column_info,
                                                                                   mixed_column_id, task)
                # Store information about converted columns
                self.target_converted_columns.update({mixed_column_id: new_type_name})

                if updated_column is not None:
                    target[:, mixed_column_id] = updated_column

        return target

    def prepare_column_types_info(self, predictors: np.array, target: np.array = None,
                                  task: Task = None) -> dict:
        """ Prepare information about columns in a form of dictionary
        Dictionary has two keys: 'target' and 'features'
        """
        if not self.features_columns_info:
            # Information about column types is empty - there is a need to launch algorithm to collect info
            self.features_columns_info = define_column_types(predictors)
            predictors = self.features_types_converting(features=predictors)
        if not self.target_columns_info and task.task_type is not TaskTypesEnum.ts_forecasting:
            self.target_columns_info = define_column_types(target)
            target = self.target_types_converting(target=target, task=task)

        features_types = _generate_list_with_types(self.features_columns_info, self.features_converted_columns)
        self._check_columns_vs_types_number(predictors, features_types)

        if target is None or task.task_type is TaskTypesEnum.ts_forecasting:
            return {'features': features_types}
        else:
            target_types = _generate_list_with_types(self.target_columns_info, self.target_converted_columns)
            self._check_columns_vs_types_number(target, target_types)
            return {'features': features_types, 'target': target_types}

    def _retain_columns_info_without_types_conflicts(self, data: InputData):
        """ Update information in supplementary info - retain info only about remained columns.
        Such columns have no conflicts with types converting.
        """
        if len(self.string_columns_transformation_failed) > 0:
            self.log.warning(f'Columns with indices {list(self.string_columns_transformation_failed.keys())} were '
                             f'removed during mixed types column converting due to conflicts.')

            data.features = self.remove_incorrect_features(data.features, self.string_columns_transformation_failed)

            remained_column_types = []
            for i, col in enumerate(data.supplementary_data.column_types['features']):
                if i not in self.string_columns_transformation_failed:
                    remained_column_types.append(col)
            data.supplementary_data.column_types['features'] = remained_column_types

    def _check_columns_vs_types_number(self, table: np.array, column_types: list):
        # Check if columns number correct
        n_rows, n_cols = table.shape
        if n_cols != len(column_types):
            # There is an incorrect types calculation
            self.log.warning('Columns number and types numbers do not match.')

    @staticmethod
    def _remove_pseudo_str_values_from_str_column(data: pd.DataFrame, column_id: int):
        """ Removes from truly str column all pseudo str values """
        cur_column = data.features[:, column_id]
        converted_column = []
        for i in range(len(cur_column)):
            try:
                float(cur_column[i])
                converted_column.append(np.nan)
            except ValueError:
                converted_column.append(cur_column[i])
        data.features[:, column_id] = pd.Series(converted_column).values

    def _convert_feature_into_one_type(self, mixed_column: np.array, column_info: dict, mixed_column_id: int):
        """ Determine new type for current feature column based on the string ratio. And then convert column into it.

        :param mixed_column: one-dimensional array with several data types
        :param column_info: dictionary with information about types in the column
        :param mixed_column_id: index of column in dataset
        """
        if len(column_info['types']) == 2 and NAME_CLASS_NONE in column_info['types']:
            # Column contain only one data type and nans
            filtered_types = [x for x in column_info['types'] if x != NAME_CLASS_NONE]
            return mixed_column, filtered_types[0]

        string_objects_number = column_info['str_number']
        all_elements_number = string_objects_number + column_info['int_number'] + column_info['float_number']
        string_ratio = string_objects_number / all_elements_number

        if string_ratio > 0:
            suggested_type = str
        else:
            suggested_type = _obtain_new_column_type(column_info)

        try:
            mixed_column = mixed_column.astype(suggested_type)
            # If there were nans in the column - paste nan
            if column_info['nan_number'] > 0:
                mixed_column = mixed_column.astype(object)
                mixed_column[column_info['nan_ids']] = np.nan
                del column_info['nan_ids']
            return mixed_column, str(suggested_type)
        except ValueError:
            # Cannot convert string objects into int or float (for example 'a' into int)
            prefix = f'Feature column with index {mixed_column_id} contains ' \
                     f'following data types: {column_info["types"]}.'
            self.log.warning(f'{prefix} String cannot be converted into {suggested_type}. Drop column.')
            return None, 'removed'

    def _convert_target_into_one_type(self, mixed_column: np.array, column_info: dict, mixed_column_id: int,
                                      task: Task) -> [np.array, str]:
        """ Convert target columns into one type based on column proportions of object and task """
        if task.task_type is TaskTypesEnum.classification:
            # For classification labels are string if at least one element is a string
            suggested_type = str
        else:
            suggested_type = _obtain_new_column_type(column_info)

        try:
            mixed_column = mixed_column.astype(suggested_type)
            return mixed_column, str(suggested_type)
        except ValueError:
            # Cannot convert string objects into int or float (for example 'a' into int)
            target_column = pd.Series(mixed_column)
            converted_column = pd.to_numeric(target_column, errors='coerce')

            prefix = f'Target column with index {mixed_column_id} contains ' \
                     f'following data types: {column_info["types"]}.'
            log_message = f'{prefix} String cannot be converted into {suggested_type}. Ignore non converted values.'
            self.log.debug(log_message)
            self.target_converting_has_errors = True
            return converted_column.values, str(suggested_type)

    def _into_categorical_features_transformation_for_fit(self, data: InputData):
        """
        Perform automated categorical features determination. If feature column
        contains int or float values with few unique values (less than 13)
        """
        n_rows, n_cols = data.features.shape
        for column_id in range(n_cols):
            # For every int/float column perform check
            column_type = data.supplementary_data.column_types['features'][column_id]
            if 'int' in column_type or 'float' in column_type:
                numerical_column = pd.Series(data.features[:, column_id])

                # Calculate number of unique values except nans
                unique_numbers = len(numerical_column.dropna().unique())

                if 2 < unique_numbers < self.categorical_max_uniques_th:
                    # Column need to be transformed into categorical (string) one
                    self.numerical_into_str.append(column_id)

                    # Convert into string
                    converted_array = convert_num_column_into_string_array(numerical_column)

                    # Store converted column into features table
                    data.features[:, column_id] = converted_array

                    # Update information about column types (in-place)
                    features_types = data.supplementary_data.column_types['features']
                    features_types[column_id] = NAME_CLASS_STR

    def _into_categorical_features_transformation_for_predict(self, data: InputData):
        """ Apply conversion into categorical string column for every signed column """
        if not self.numerical_into_str:
            # There is no transformation for current table
            return data

        n_rows, n_cols = data.features.shape
        for column_id in range(n_cols):
            if column_id in self.numerical_into_str:
                numerical_column = pd.Series(data.features[:, column_id])
                # Column must be converted into categorical
                converted_array = convert_num_column_into_string_array(numerical_column)
                data.features[:, column_id] = converted_array

                # Update information about column types (in-place)
                features_types = data.supplementary_data.column_types['features']
                features_types[column_id] = NAME_CLASS_STR

    def _into_numeric_features_transformation_for_fit(self, data: InputData):
        """
        Automatically determine categorical features which should be converted into float
        """
        n_rows, n_cols = data.features.shape
        for column_id in range(n_cols):
            # For every string column perform converting if necessary
            column_type = data.supplementary_data.column_types['features'][column_id]
            if 'str' in column_type:
                string_column = pd.Series(data.features[:, column_id])

                # Number of nans in the column
                nans_number = string_column.isna().sum()

                # Column probably not an "actually categorical" but a column with an incorrectly defined type
                converted_column = pd.to_numeric(string_column, errors='coerce')
                # Calculate applied nans
                result_nans_number = converted_column.isna().sum()
                failed_objects_number = result_nans_number - nans_number
                non_nan_all_objects_number = n_rows - nans_number
                failed_ratio = failed_objects_number / non_nan_all_objects_number

                # If all objects are truly strings - all objects transform into nan
                is_column_contain_numerical_objects = failed_ratio != 1
                if failed_ratio < self.acceptable_failed_rate_bottom:
                    # The majority of objects can be converted into numerical
                    data.features[:, column_id] = converted_column.values

                    # Update information about column types (in-place)
                    self.categorical_into_float.append(column_id)
                    features_types = data.supplementary_data.column_types['features']
                    features_types[column_id] = NAME_CLASS_FLOAT
                elif failed_ratio >= self.acceptable_failed_rate_top \
                        and is_column_contain_numerical_objects:
                    # The column consists mostly of truly str values and has a few ints/floats in it
                    self._remove_pseudo_str_values_from_str_column(data, column_id)
                elif self.acceptable_failed_rate_top > failed_ratio >= self.acceptable_failed_rate_bottom:
                    # Probably numerical column contains a lot of '?' or 'x' as nans equivalents
                    # Add columns to remove list
                    self.string_columns_transformation_failed.update({column_id: 'removed'})

    def _into_numeric_features_transformation_for_predict(self, data: InputData):
        """ Apply conversion into float string column for every signed column """
        if not self.categorical_into_float:
            # There is no transformation for current table
            return data

        n_rows, n_cols = data.features.shape
        for column_id in range(n_cols):
            if column_id in self.categorical_into_float and column_id not in self.string_columns_transformation_failed:
                string_column = pd.Series(data.features[:, column_id])

                # Column must be converted into float from categorical
                converted_column = pd.to_numeric(string_column, errors='coerce')
                data.features[:, column_id] = converted_column.values

                # Update information about column types (in-place)
                features_types = data.supplementary_data.column_types['features']
                features_types[column_id] = NAME_CLASS_FLOAT


def define_column_types(table: np.array):
    """ Prepare information about types per columns. For each column store unique
    types, which column contains. If column with mixed type contain str object
    additional field 'str_ids' with indices of string objects is prepared
    """

    # TODO: current processing is relatively computationally expensive - probably refactor needed

    def type_ignoring_nans(item):
        """ Return type of element in the array. If item is np.nan - return NoneType """
        current_type = type(item)
        if current_type is float and np.isnan(item):
            # Check is current element is nan or not (np.nan is a float type)
            return type(None)
        return current_type

    if table is None:
        return {}

    n_rows, n_columns = table.shape
    columns_info = {}
    for column_id in range(n_columns):
        current_column = table[:, column_id]

        # Check every element in numpy array - it can take a long time!
        column_types = list(map(type_ignoring_nans, current_column))

        # Store only unique values
        set_column_types = set(column_types)
        # Convert types into string names
        column_types_names = list(map(str, set_column_types))

        if len(column_types_names) > 1:
            # There are several types in one column
            types_names = np.array(column_types, dtype=str)
            # Calculate number of string objects in the dataset
            str_number = len(np.argwhere(types_names == NAME_CLASS_STR))
            int_number = len(np.argwhere(types_names == NAME_CLASS_INT))
            float_number = len(np.argwhere(types_names == NAME_CLASS_FLOAT))

            # Store information about nans in the target
            nan_ids = np.ravel(np.argwhere(types_names == NAME_CLASS_NONE))
            nan_number = len(nan_ids)
            columns_info.update({column_id: {'types': column_types_names,
                                             'str_number': str_number,
                                             'int_number': int_number,
                                             'float_number': float_number,
                                             'nan_number': nan_number,
                                             'nan_ids': nan_ids}})
        else:
            # There is only one type, or several types such as int and float
            columns_info.update({column_id: {'types': column_types_names}})
    return columns_info


def find_mixed_types_columns(columns_info: dict):
    """ Search for columns with several types in them """
    columns_with_mixed_types = []
    for column_id, information in columns_info.items():
        column_types = information['types']
        if len(column_types) > 1:
            columns_with_mixed_types.append(column_id)

    return columns_with_mixed_types


def apply_type_transformation(table: np.array, column_types: list, log: LoggerAdapter):
    """
    Apply transformation for columns in dataset into desired type. Perform
    transformation on predict stage when column types were already determined
    during fit
    """

    def type_by_name(current_type_name: str):
        """ Return type by its name """
        if 'int' in current_type_name:
            return int
        elif 'str' in current_type_name:
            return str
        else:
            return float

    if table is None:
        # Occurs if for predict stage there is no target info
        return None

    n_rows, n_cols = table.shape
    for column_id in range(n_cols):
        current_column = table[:, column_id]
        current_type = type_by_name(column_types[column_id])
        _convert_predict_column_into_desired_type(table=table, current_column=current_column, current_type=current_type,
                                                  column_id=column_id, log=log)

    return table


def convert_num_column_into_string_array(numerical_column: pd.Series) -> np.array:
    """ Convert pandas column into numpy one-dimensional array """
    # Convert into string
    converted_column = numerical_column.astype(str)
    converted_array = converted_column.values

    # If there are nans - insert them
    nan_ids = np.ravel(np.argwhere(converted_array == 'nan'))
    if len(nan_ids) > 0:
        converted_array = converted_array.astype(object)
        converted_array[nan_ids] = np.nan

    return converted_array


def _obtain_new_column_type(column_info):
    """ Suggest in or float type based on the presence of nan and float values """
    if column_info['float_number'] > 0 or column_info['nan_number'] > 0:
        # Even if one of types are float - all elements should be converted into float
        return float
    else:
        # It is available to convert numerical into integer type
        return int


def _convert_predict_column_into_desired_type(table: np.array, current_column: np.array,
                                              column_id: int, current_type, log: LoggerAdapter):
    try:
        table[:, column_id] = current_column.astype(current_type)
        if current_type is str:
            is_any_comma = any(map(lambda el: ',' in el, current_column))
            is_any_dot = any(map(lambda el: '.' in el, current_column))
            # Most likely case: '20,000' must be converted into '20.000'
            if is_any_comma and is_any_dot:
                warning = f'Column {column_id} contains both "." and ",". Standardize it.'
                log.warning(warning)
    except ValueError:
        table[:, column_id] = _process_predict_column_values_one_by_one(current_column=current_column,
                                                                        current_type=current_type)


def _generate_list_with_types(columns_types_info: dict, converted_columns: dict) -> list:
    """ Create list with types for all remained columns

    :param columns_types_info: dictionary with initial column types
    :param converted_columns: dictionary with transformed column types
    """
    updated_column_types = []
    for column_id, column_info in columns_types_info.items():
        column_types = column_info['types']

        if len(column_types) == 1:
            # Column initially contain only one type
            updated_column_types.append(column_types[0])
        elif len(column_types) == 2 and NAME_CLASS_NONE in column_types:
            # Column with one type and nans
            filtered_types = [x for x in column_types if x != NAME_CLASS_NONE]
            updated_column_types.append(filtered_types[0])
        else:
            if any('str' in column_type_name for column_type_name in column_types):
                # Mixed-types column with string
                new_column_type = converted_columns[column_id]
                if new_column_type != 'removed':
                    updated_column_types.append(new_column_type)
            else:
                # Mixed-types with float and integer
                updated_column_types.append(NAME_CLASS_FLOAT)

    return updated_column_types


def _process_predict_column_values_one_by_one(current_column: np.ndarray, current_type):
    """ Process column values one by one and try to convert them into desirable type.
    If not successful replace with np.nan """

    def _process_str_numbers_with_dots_and_commas(value: str):
        """ Try to process str with replacing ',' by '.' in case it was meant to be a number """
        value = value.replace(',', '.')
        try:
            # Since "10.6" can not be converted to 10 straightforward using int()
            if current_type is int:
                new_value = int(float(value))
            else:
                new_value = current_type(value)
        except ValueError:
            return np.nan
        return new_value

    new_column = []
    for value in current_column:
        new_value = np.nan
        try:
            new_value = current_type(value)
        except ValueError:
            if isinstance(value, str) and ('.' in value or ',' in value):
                new_value = _process_str_numbers_with_dots_and_commas(value=value)
        new_column.append(new_value)
    return new_column



import json
from abc import abstractmethod
from os.path import join
from typing import List, Union

from fedot.core.utils import fedot_project_root


def load_params_bounds_file():
    params_bounds_file = join(fedot_project_root(), 'fedot', 'sensitivity',
                              'operations_hp_sensitivity', 'params_bounds.json')
    with open(params_bounds_file, 'r') as file:
        data = json.load(file)

    return data


class Problem:
    operation_params_with_bounds_by_operation_name = load_params_bounds_file()

    INTEGER_PARAMS = ['n_estimators', 'n_neighbors', 'p', 'min_child_weight', 'max_depth', 'n_clusters',
                      'min_samples_split', 'min_samples_leaf', 'd',
                      'q', 'lag_1', 'lag_2', 'max_trials', 'max_skips',
                      'degree', 'window_size', 'sigma']

    def __init__(self, operation_types: Union[List[str], str]):
        self.operation_types = operation_types
        self.num_vars = None
        self.names = None
        self.bounds = None

    @abstractmethod
    def convert_sample_to_dict(self, samples) -> Union[List[dict], List[List[dict]]]:
        raise NotImplementedError

    def clean_sample_variables(self, samples: List[dict]):
        """Make integer values for params if necessary"""
        for sample in samples:
            if sample is not None:
                for key, value in sample.items():
                    if key in self.INTEGER_PARAMS:
                        sample[key] = int(value)

        return samples

    @property
    def dictionary(self):
        problem = {
            'num_vars': self.num_vars,
            'names': self.names,
            'bounds': self.bounds
        }

        return problem


class OneOperationProblem(Problem):

    def __init__(self, operation_types: Union[List[str], str]):
        super().__init__(operation_types)
        self.params: dict = \
            self.operation_params_with_bounds_by_operation_name.get(self.operation_types[0])
        self.num_vars: int = len(self.params)
        self.names: List[str] = list(self.params.keys())
        self.bounds: List[list] = list(self.params.values())

    def convert_sample_to_dict(self, samples) -> List[dict]:
        converted_samples = []
        for sample in samples:
            new_params = {}
            for index, value in enumerate(sample):
                new_params[self.names[index]] = value
            converted_samples.append(new_params)

        converted_and_cleaned_samples = self.clean_sample_variables(converted_samples)

        return converted_and_cleaned_samples

    def convert_for_dispersion_analysis(self, transposed_samples) -> List[List[dict]]:
        """
        convert samples into dicts per param as follows:
        [
            [{'p1': 'v1'},{'p1': 'v2'},...{'p1': 'vn'},],
            [{'p2': 'v1'},{'p2': 'v2'},...{'p2': 'vn'},],
            ...
            [{'pm': 'v1'},{'pm': 'v2'},...{'pm': 'vn'},]]

        :param transposed_samples:
        :return:
        """

        converted_samples: List[list] = []
        for index, param in enumerate(self.names):
            samples_per_param = [{param: value} for value in transposed_samples[index]]
            cleaned = self.clean_sample_variables(samples_per_param)
            converted_samples.append(cleaned)

        return converted_samples


class MultiOperationsProblem(Problem):
    def __init__(self, operation_types: List[str]):
        super().__init__(operation_types)
        self.params: List[dict] = \
            [self.operation_params_with_bounds_by_operation_name.get(operation_type)
             for operation_type in self.operation_types]
        self.num_vars: int = sum([len(operation_params) for operation_params in self.params
                                  if operation_params is not None])
        self.names: List[str] = list()
        self.names_per_node: dict = dict()
        self.bounds: List[list] = list()
        for index, operation_params in enumerate(self.params):
            if operation_params is not None:
                self.names.extend(list(operation_params.keys()))
                self.names_per_node[str(index)] = list(operation_params.keys())
                self.bounds.extend(list(operation_params.values()))
            else:
                self.names_per_node[str(index)] = None

    def convert_sample_to_dict(self, samples) -> Union[List[dict], List[List[dict]]]:
        all_operations_new_params: List[List[dict]] = list()
        for sample in samples:
            new_params_per_operation: List[Union[dict, None]] = list()
            border_index = 0
            for _, params_names in self.names_per_node.items():
                if params_names is not None:
                    new_params_values_per_operation = \
                        sample[border_index:border_index + len(params_names)]
                    node_params = dict(zip(params_names, new_params_values_per_operation))
                    new_params_per_operation.append(node_params)
                    border_index += len(params_names)
                else:
                    new_params_per_operation.append(None)

            cleaned_new_params_per_params = self.clean_sample_variables(new_params_per_operation)
            all_operations_new_params.append(cleaned_new_params_per_params)

        return all_operations_new_params

        # Apply new params to nodes in pipeline in operation sensitivity



from matplotlib import pyplot as plt

from fedot.core.data.data import InputData
from fedot.core.pipelines.ts_wrappers import fitted_values, in_sample_fitted_values
from fedot.core.repository.tasks import Task, TaskTypesEnum, TsForecastingParams
from test.unit.pipelines.test_pipeline_ts_wrappers import get_simple_short_lagged_pipeline


def show_fitted_time_series(len_forecast=24):
    """
    Shows an example of how to get fitted values of a time series by any
    pipeline created by FEDOT

    fitted values - are the predictions of the pipelines on the training sample.
    For time series, these values show how well the model reproduces the time
    series structure
    """
    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=len_forecast))

    ts_input = InputData.from_csv_time_series(file_path='../../../cases/data/time_series/metocean.csv',
                                              task=task, target_column='value')

    pipeline = get_simple_short_lagged_pipeline()
    train_predicted = pipeline.fit(ts_input)

    # Get fitted values for every 10th forecast
    fitted_ts_10 = fitted_values(ts_input, train_predicted, 10)
    # Average for all forecasting horizons
    fitted_ts_act = fitted_values(ts_input, train_predicted)
    # In-sample forecasting fitted values
    in_sample_validated = in_sample_fitted_values(ts_input, train_predicted)

    plt.plot(range(len(ts_input.idx)), ts_input.target, label='Actual time series', alpha=0.8)
    plt.plot(fitted_ts_10.idx, fitted_ts_10.predict, label='Fitted values horizon 10', alpha=0.2)
    plt.plot(fitted_ts_act.idx, fitted_ts_act.predict, label='Fitted values all', alpha=0.2)
    plt.plot(in_sample_validated.idx, in_sample_validated.predict, label='In-sample fitted values')
    plt.legend()
    plt.grid()
    plt.show()


if __name__ == '__main__':
    show_fitted_time_series()



import warnings
from typing import Optional

from fedot.core.data.data import InputData, OutputData
from fedot.core.operations.evaluation.evaluation_interfaces import EvaluationStrategy
from fedot.core.operations.evaluation.operation_implementations.data_operations.categorical_encoders import \
    OneHotEncodingImplementation, LabelEncodingImplementation
from fedot.core.operations.evaluation.operation_implementations.data_operations.sklearn_transformations import \
    ImputationImplementation, KernelPCAImplementation, NormalizationImplementation, PCAImplementation, \
    PolyFeaturesImplementation, ScalingImplementation, FastICAImplementation
from fedot.core.operations.operation_parameters import OperationParameters
from fedot.core.utilities.random import RandomStateHandler

warnings.filterwarnings("ignore", category=UserWarning)


class FedotPreprocessingStrategy(EvaluationStrategy):
    """
    Args:
        operation_type: ``str`` of the operation defined in operation or data operation repositories

            .. details:: possible operations:

                - ``scaling``-> ScalingImplementation,
                - ``normalization``-> NormalizationImplementation,
                - ``simple_imputation``-> ImputationImplementation,
                - ``pca``-> PCAImplementation,
                - ``kernel_pca``-> KernelPCAImplementation,
                - ``poly_features``-> PolyFeaturesImplementation,
                - ``one_hot_encoding``-> OneHotEncodingImplementation,
                - ``label_encoding``-> LabelEncodingImplementation,
                - ``fast_ica``-> FastICAImplementation

        params: hyperparameters to fit the operation with

    """

    __operations_by_types = {
        'scaling': ScalingImplementation,
        'normalization': NormalizationImplementation,
        'simple_imputation': ImputationImplementation,
        'pca': PCAImplementation,
        'kernel_pca': KernelPCAImplementation,
        'poly_features': PolyFeaturesImplementation,
        'one_hot_encoding': OneHotEncodingImplementation,
        'label_encoding': LabelEncodingImplementation,
        'fast_ica': FastICAImplementation
    }

    def __init__(self, operation_type: str, params: Optional[OperationParameters] = None):
        self.operation_impl = self._convert_to_operation(operation_type)
        super().__init__(operation_type, params)

    def fit(self, train_data: InputData):
        """This method is used for operation training with the data provided

        Args:
            train_data: data used for operation training

        Returns:
            trained Sklearn operation
        """

        warnings.filterwarnings("ignore", category=RuntimeWarning)
        operation_implementation = self.operation_impl(self.params_for_fit)
        with RandomStateHandler():
            operation_implementation.fit(train_data)
        return operation_implementation

    def predict(self, trained_operation, predict_data: InputData) -> OutputData:
        """Transform method for preprocessing task

        Args:
            trained_operation: model object
            predict_data: data used for prediction

        Returns:
            prediction
        """
        prediction = trained_operation.transform(predict_data)
        # Convert prediction to output (if it is required)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def predict_for_fit(self, trained_operation, predict_data: InputData) -> OutputData:
        """
        Transform method for preprocessing task for fit stage

        Args:
            trained_operation: model object
            predict_data: data used for prediction
        Returns:
            OutputData: 
        """
        prediction = trained_operation.transform_for_fit(predict_data)
        converted = self._convert_to_output(prediction, predict_data)
        return converted

    def _convert_to_operation(self, operation_type: str):
        if operation_type in self.__operations_by_types.keys():
            return self.__operations_by_types[operation_type]
        else:
            raise ValueError(f'Impossible to obtain custom preprocessing strategy for {operation_type}')



import warnings

import numpy as np
from matplotlib import pyplot as plt
from sklearn.metrics import mean_absolute_error

from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup

warnings.filterwarnings('ignore')


def plot_averaged_forecast(actual, predicted, first_column_act, last_column_act,
                           first_column_pr, last_column_pr):
    """
    Function plot averaged forecasts and actual time series

    :param actual: time series with averaged actual values
    :param predicted: time series with averaged predicted values
    :param first_column_act: numpy array with actual first days
    :param last_column_act: numpy array with actual seventh days
    :param first_column_pr: numpy array with predicted first days
    :param last_column_pr: numpy array with predicted seventh days
    """
    plt.plot(actual, label='7-day moving average actual')
    plt.fill_between(range(0, len(actual)), first_column_act, last_column_act, alpha=0.4)
    plt.plot(predicted, label='7-day moving average forecast')
    plt.fill_between(range(0, len(actual)), first_column_pr, last_column_pr, alpha=0.4)
    plt.ylabel('River level, cm', fontsize=14)
    plt.xlabel('Time index', fontsize=14)
    plt.grid()
    plt.legend(fontsize=12)
    plt.show()


def plot_last_day_forecast(last_column_act, last_column_pr):
    """ The function draws model predictions and actual water level values for
    every seventh day of the forecast (forecast horizon is 7 days)

    :param last_column_act: numpy array with actual seventh days
    :param last_column_pr: numpy array with predicted seventh days
    """
    plt.plot(last_column_act, label='7th day actual')
    plt.plot(last_column_pr, label='7th day forecast')
    plt.ylabel('River level, cm', fontsize=14)
    plt.xlabel('Time index', fontsize=14)
    plt.grid()
    plt.legend(fontsize=12)
    plt.show()


def plot_predictions(predicted_array, test_output):
    """ Function plot the predictions of the algorithm """
    predicted_columns = np.array(predicted_array)
    actual_columns = np.array(test_output.target)

    # Take mean value for columns and also first and last columns
    predicted = predicted_columns.mean(axis=1)
    first_column_pr = np.ravel(predicted_columns[:, 0])
    last_column_pr = np.ravel(predicted_columns[:, -1])
    actual = actual_columns.mean(axis=1)
    first_column_act = np.ravel(actual_columns[:, 0])
    last_column_act = np.ravel(actual_columns[:, -1])

    plot_averaged_forecast(actual, predicted, first_column_act,
                           last_column_act,
                           first_column_pr, last_column_pr)

    plot_last_day_forecast(last_column_act, last_column_pr)


def run_multi_output_case(path, vis=False):
    """ Function launch case for river levels prediction on Lena river as
    multi-output regression task

    :param path: path to the file with table
    :param vis: is it needed to visualise pipeline and predictions
    """
    target_columns = ['1_day', '2_day', '3_day', '4_day', '5_day', '6_day', '7_day']

    data = InputData.from_csv(path, target_columns=target_columns, columns_to_drop=['date'])
    train, test = train_test_data_setup(data)

    problem = 'regression'

    automl_model = Fedot(problem=problem)
    automl_model.fit(features=train)
    predicted_array = automl_model.predict(features=test)

    # Convert output into one dimensional array
    forecast = np.ravel(predicted_array)

    mae_value = mean_absolute_error(np.ravel(test.target), forecast)
    print(f'MAE - {mae_value:.2f}')

    if vis:
        plot_predictions(predicted_array, test)


if __name__ == '__main__':
    path_file = './data/lena_levels/multi_sample.csv'
    run_multi_output_case(path_file, vis=True)



from typing import Optional, Any

import matplotlib.pyplot as plt
import numpy as np

from fedot.core.composer.metrics import ROCAUC
from fedot.core.data.data import InputData, OutputData
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.repository.dataset_types import DataTypesEnum


def plot_forecast(data: [InputData, MultiModalData], prediction: OutputData, in_sample: bool = False,
                  target: Optional[Any] = None):
    """
    Function for drawing plot with time series forecast. If data.target is None function plot prediction
    as future values. If not - we use last data features as validation.

    Args:
        data: the InputData or MultiModalData with actual time series as features
        prediction: the OutputData with predictions
        in_sample: if obtained prediction was in sample.
        If ``False`` plots predictions as future values for test data features.
        target: user-specified name of target variable for MultiModalData
    """
    if isinstance(data, MultiModalData):
        if not target:
            raise AttributeError("Can't visualize. Target of MultiModalData not set.")
        data = data.extract_data_source(target)
    if data.data_type == DataTypesEnum.multi_ts:
        actual_time_series = data.features[:, 0]
    else:
        actual_time_series = data.features
    target_time_series = data.target
    predict = prediction.predict
    pred_start = len(actual_time_series)
    if in_sample:
        pred_start = len(actual_time_series) - len(predict)
    elif target_time_series is not None:
        actual_time_series = np.concatenate([actual_time_series, target_time_series], axis=0)

    padding = min(pred_start, 72)

    first_idx = pred_start - padding

    plt.plot(np.arange(pred_start, pred_start + len(predict)),
             predict, label='Predicted', c='blue')
    plt.plot(np.arange(first_idx, len(actual_time_series)),
             actual_time_series[first_idx:], label='Actual values', c='green')

    # Plot black line which divide our array into train and test
    plt.plot([pred_start, pred_start],
             [min(actual_time_series[first_idx:]), max(actual_time_series[first_idx:])], c='black',
             linewidth=1)
    plt.legend(fontsize=15)
    plt.grid()
    plt.show()


def plot_biplot(prediction: OutputData):
    """
    Function for drawing biplot with regression task.

    :param prediction: the OutputData with prediction and target.
    Target should be not null
    """

    target = prediction.target
    predict = prediction.predict

    if target is None:
        raise ValueError('Target must be not None to plot biplot')

    plt.figure(figsize=(10, 10))
    plt.scatter(target, predict)
    plt.grid()
    # plot bisect
    min_coord = min(np.min(target), np.min(predict))
    max_coord = max(np.max(target), np.max(predict))
    min_max_bisect = [min_coord, max_coord]
    plt.plot(min_max_bisect, min_max_bisect, c='black', linewidth=1)

    plt.title("Biplot")
    plt.xlabel("target")
    plt.ylabel("prediction")
    plt.show()


def plot_roc_auc(data: InputData, prediction: OutputData):
    """
    Function for drawing roc curve for classification task.

    :param data: the InputData with validation data
    :param prediction: prediction for data
    """
    if data.num_classes == 2:
        fpr, tpr, threshold = ROCAUC.roc_curve(data.target, prediction.predict)
        roc_auc = ROCAUC.auc(fpr, tpr)
        plt.plot(fpr, tpr, 'b', label='AUC = %0.2f' % roc_auc)
        plt.legend(loc=F'best')
    else:
        for i in range(data.num_classes):
            fpr, tpr, threshold = ROCAUC.roc_curve(data.target, prediction.predict[:, i],
                                                   pos_label=data.class_labels[i])
            roc_auc = ROCAUC.auc(fpr, tpr)
            plt.plot(fpr, tpr, label=f'label-{i} AUC = %0.2f' % roc_auc)
            plt.legend(loc=F'best')
    plt.plot([0, 1], [0, 1], 'r--')
    plt.xlim([0, 1])
    plt.ylim([0, 1])
    plt.ylabel('True Positive Rate')
    plt.xlabel('False Positive Rate')
    plt.show()



from functools import partial
from typing import Tuple, Optional

from hyperopt import fmin, space_eval, Trials

from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.tuning.search_space import convert_params, get_node_operation_parameter_label
from fedot.core.pipelines.tuning.tuner_interface import HyperoptTuner


class PipelineTuner(HyperoptTuner):
    """
    Class for hyperparameters optimization for all nodes simultaneously
    """

    def tune(self, pipeline: Pipeline, show_progress: bool = True) -> Pipeline:
        """ Function for hyperparameters tuning on the entire pipeline

        :param pipeline: Pipeline which hyperparameters will be tuned
        :param show_progress: shows progress of tuning if true
        """

        parameters_dict, init_parameters = self._get_parameters_for_tune(pipeline)

        self.init_check(pipeline)

        pipeline.replace_n_jobs_in_nodes(n_jobs=self.n_jobs)

        trials = Trials()

        # try searching using initial parameters (uses original search space with fixed initial parameters)
        trials, init_trials_num = self._search_near_initial_parameters(pipeline, parameters_dict, init_parameters,
                                                                       trials, show_progress)

        best = fmin(partial(self._objective, pipeline=pipeline),
                    parameters_dict,
                    trials=trials,
                    algo=self.algo,
                    max_evals=self.iterations,
                    show_progressbar=show_progress,
                    early_stop_fn=self.early_stop_fn,
                    timeout=self.max_seconds)

        best = space_eval(space=parameters_dict, hp_assignment=best)
        # check if best point was obtained using search space with fixed initial parameters
        is_best_trial_with_init_params = trials.best_trial.get('tid') in range(init_trials_num)
        if is_best_trial_with_init_params:
            best = {**best, **init_parameters}

        tuned_pipeline = self.set_arg_pipeline(pipeline=pipeline,
                                               parameters=best)

        # Validation is the optimization do well
        final_pipeline = self.final_check(tuned_pipeline)

        return final_pipeline

    def _search_near_initial_parameters(self, pipeline: Pipeline, search_space: dict, initial_parameters: dict,
                                        trials: Trials, show_progress: bool = True):
        try_initial_parameters = initial_parameters and self.iterations > 1
        if not try_initial_parameters:
            init_trials_num = 0
            return trials, init_trials_num

        is_init_params_full = len(initial_parameters) == len(search_space)
        if self.iterations < 10 or is_init_params_full:
            init_trials_num = 1
        else:
            init_trials_num = min(int(self.iterations * 0.1), 10)

        # fmin updates trials with evaluation points tried out during the call
        fmin(partial(self._objective, pipeline=pipeline, unchangeable_parameters=initial_parameters),
             search_space,
             trials=trials,
             algo=self.algo,
             max_evals=init_trials_num,
             show_progressbar=show_progress,
             early_stop_fn=self.early_stop_fn,
             timeout=self.max_seconds)
        return trials, init_trials_num

    def _get_parameters_for_tune(self, pipeline: Pipeline) -> Tuple[dict, dict]:
        """
        Function for defining the search space

        :return parameters_dict: dictionary with operation names and parameters
        """

        parameters_dict = {}
        initial_parameters = {}
        for node_id, node in enumerate(pipeline.nodes):
            operation_name = node.operation.operation_type

            # Assign unique prefix for each model hyperparameter
            # label - number of node in the pipeline
            node_params = self.search_space.get_node_params(node_id=node_id,
                                                            operation_name=operation_name)

            if node_params is not None:
                parameters_dict.update(node_params)

            tunable_node_params = self.search_space.get_operation_parameter_range(operation_name)
            tunable_initial_params = {get_node_operation_parameter_label(node_id, operation_name, p):
                                      node.parameters[p] for p in node.parameters if p in tunable_node_params}
            if tunable_initial_params:
                initial_parameters.update(tunable_initial_params)

        return parameters_dict, initial_parameters

    def _objective(self, parameters_dict: dict, pipeline: Pipeline, unchangeable_parameters: Optional[dict] = None) \
            -> float:
        """
        Objective function for minimization / maximization problem

        Args:
            parameters_dict: Dict which contains new pipeline hyperparameters
            pipeline: pipeline to optimize
            unchangeable_parameters: Dict with parameters that should not be changed

        Returns:
             metric_value: value of objective function
        """

        # replace new parameters with parameters
        if unchangeable_parameters:
            parameters_dict = {**parameters_dict, **unchangeable_parameters}
        # Set hyperparameters for every node
        pipeline = self.set_arg_pipeline(pipeline, parameters_dict)
        metric_value = self.get_metric_value(pipeline=pipeline)
        return metric_value

    @staticmethod
    def set_arg_pipeline(pipeline: Pipeline, parameters: dict) -> Pipeline:
        """ Method for parameters setting to a pipeline

        Args:
            pipeline: pipeline to which parameters should ba assigned
            parameters: dictionary with parameters to set

        Returns:
            pipeline: pipeline with new hyperparameters in each node
        """
        # Set hyperparameters for every node
        for node_id, node in enumerate(pipeline.nodes):
            node_params = {key: value for key, value in parameters.items()
                           if key.startswith(f'{str(node_id)} || {node.name}')}

            if node_params is not None:
                # Delete all prefix strings to get appropriate parameters names
                new_params = convert_params(node_params)

                # Update parameters in nodes
                pipeline.nodes[node_id].parameters = new_params

        return pipeline



import glob
import json
import os
from abc import ABC, abstractmethod

import pandas as pd


class BatchLoader(ABC):
    """ Class for loading data with batches """

    def __init__(self, path: str, label='label'):
        self.path = path
        self.meta_df = None
        self.target_name = label

    @abstractmethod
    def extract(self) -> pd.DataFrame:
        pass

    def _extract_files_paths(self):
        all_files = []
        for root, dirs, files in os.walk(self.path):
            files_paths = []
            for name in files:
                if not name.startswith('.'):
                    whole_file_path = os.path.join(root, name)
                    files_paths.append(whole_file_path)

            if files:
                all_files.extend(files_paths)
        self._load_to_meta_df(all_files)

    def _load_to_meta_df(self, files, shuffle=True):
        data_rows = []
        for file in files:
            dir_name = os.path.basename(os.path.dirname(file))
            row = [file, dir_name]
            data_rows.append(row)

        self.meta_df = pd.DataFrame(data=data_rows,
                                    columns=['file_path', f'{self.target_name}'])
        # shuffle samples
        if shuffle:
            self.meta_df = self.meta_df.sample(frac=1).reset_index(drop=True)

    def export_to_csv(self, path: str = None):
        if not path:
            path = self.path
            export_filename = f'meta_{os.path.basename(path)}.csv'
            export_dirname = os.path.dirname(path)
            self.meta_df.to_csv(os.path.join(export_dirname, export_filename))
        else:
            self.meta_df.to_csv(os.path.abspath(path))


class TextBatchLoader(BatchLoader):
    """ Class for loading text data with batches """

    def __init__(self, path: str):
        if os.path.isfile(path):
            raise ValueError('Expected directory path but got file')
        super().__init__(path)

    def extract(self, export: bool = True):
        self._extract_files_paths()

        content_list = []
        for file_path in self.meta_df['file_path'].tolist():
            with open(file_path, 'r') as text_file:
                content = text_file.read()
                content_list.append(content)

        new_column = pd.Series(data=content_list, index=self.meta_df.index)

        self.meta_df.insert(loc=self.meta_df.columns.get_loc('file_path'),
                            column='text',
                            value=new_column)
        self.meta_df = self.meta_df.drop(['file_path'], axis=1)

        if export:
            self.export_to_csv()

        return self.meta_df


class JSONBatchLoader(BatchLoader):
    """ Class for loading json data with batches """

    def __init__(self, path: str, label: str, fields_to_use: list, shuffle=True):
        self.fields_to_use = fields_to_use
        self.shuffle = shuffle
        if os.path.isfile(path):
            raise ValueError('Expected directory path but got file')
        super().__init__(path, label)

    def extract(self, export: bool = False):
        self._extract_files_paths(extension='json')

        dict_list = []
        labels = []
        for file_path in self.meta_df['file_path'].tolist():
            with open(file_path, 'r') as text_file:
                content = text_file.read()
                content_dict = json.loads(content)
                dict_list.append(content_dict)
                try:
                    label = content_dict[self.target_name]
                except KeyError:
                    print(file_path)
                    label = None
                labels.append(label)

        self.meta_df[self.target_name] = pd.Series(data=labels, index=self.meta_df.index)

        for field in self.fields_to_use:
            # generate feature column for each extracted field
            new_feature = pd.Series(data=[d[field] for d in dict_list], index=self.meta_df.index)
            # add column with name 'field' and value 'new_feature'
            # to the features data frame after 'file_path' column
            self.meta_df.insert(loc=self.meta_df.columns.get_loc('file_path'),
                                column=field, value=new_feature)
        # remove redunant column with file path
        self.meta_df = self.meta_df.drop(['file_path'], axis=1)

        if export:
            self.export_to_csv()

        return self.meta_df

    def _extract_files_paths(self, extension: str = ''):
        all_files = []
        path = os.path.join(self.path, f'*.{extension}') if extension else self.path
        for files in glob.glob(path):
            files_paths = []
            name = os.path.basename(files)
            if not name.startswith('.'):
                files_paths.append(files)

            if files:
                all_files.extend(files_paths)
        self._load_to_meta_df(all_files, self.shuffle)



from datetime import timedelta

import numpy as np
from sklearn.metrics import mean_absolute_error
from sklearn.model_selection import train_test_split

from examples.simple.regression.regression_pipelines import regression_ransac_pipeline
from fedot.core.data.data import InputData
from fedot.core.pipelines.tuning.sequential import SequentialTuner
from fedot.core.pipelines.tuning.tuner_builder import TunerBuilder
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.quality_metrics_repository import RegressionMetricsEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum
from fedot.utilities.synth_dataset_generator import regression_dataset

np.random.seed(2020)


def get_regression_dataset(features_options, samples_amount=250,
                           features_amount=5):
    """
    Prepares four numpy arrays with different scale features and target
    :param samples_amount: Total amount of samples in the resulted dataset.
    :param features_amount: Total amount of features per sample.
    :param features_options: The dictionary containing features options in key-value
    format:
        - informative: the amount of informative features;
        - bias: bias term in the underlying linear model;
    :return x_data_train: features to train
    :return y_data_train: target to train
    :return x_data_test: features to test
    :return y_data_test: target to test
    """

    x_data, y_data = regression_dataset(samples_amount=samples_amount,
                                        features_amount=features_amount,
                                        features_options=features_options,
                                        n_targets=1,
                                        noise=0.0, shuffle=True)

    # Changing the scale of the data
    for i, coeff in zip(range(0, features_amount),
                        np.random.randint(1, 100, features_amount)):
        # Get column
        feature = np.array(x_data[:, i])

        # Change scale for this feature
        rescaled = feature * coeff
        x_data[:, i] = rescaled

    # Train and test split
    x_train, x_test, y_train, y_test = train_test_split(x_data, y_data,
                                                        test_size=0.3)

    return x_train, y_train, x_test, y_test


def run_experiment(pipeline, tuner):
    samples = [50, 250, 150]
    features = [1, 5, 10]
    options = [{'informative': 1, 'bias': 0.0},
               {'informative': 2, 'bias': 2.0},
               {'informative': 1, 'bias': 3.0}]

    for samples_amount, features_amount, features_options in zip(samples, features, options):
        print('=======================================')
        print(f'\nAmount of samples {samples_amount}, '
              f'amount of features {features_amount}, '
              f'additional options {features_options}')

        x_train, y_train, x_test, y_test = get_regression_dataset(features_options,
                                                                  samples_amount,
                                                                  features_amount)

        # Define regression task
        task = Task(TaskTypesEnum.regression)

        # Prepare data to train the model
        train_input = InputData(idx=np.arange(0, len(x_train)),
                                features=x_train,
                                target=y_train,
                                task=task,
                                data_type=DataTypesEnum.table)

        predict_input = InputData(idx=np.arange(0, len(x_test)),
                                  features=x_test,
                                  target=None,
                                  task=task,
                                  data_type=DataTypesEnum.table)

        # Fit it
        pipeline.fit(train_input)

        # Predict
        predicted_values = pipeline.predict(predict_input)
        pipeline_prediction = predicted_values.predict

        mae_value = mean_absolute_error(y_test, pipeline_prediction)
        print(f'Mean absolute error - {mae_value:.4f}\n')

        if tuner is not None:
            print('Start tuning process ...')
            pipeline_tuner = TunerBuilder(task)\
                .with_tuner(tuner)\
                .with_metric(RegressionMetricsEnum.MAE)\
                .with_iterations(50) \
                .with_timeout(timedelta(seconds=50))\
                .build(train_input)
            tuned_pipeline = pipeline_tuner.tune(pipeline)

            # Fit it
            tuned_pipeline.fit(train_input)

            # Predict
            predicted_values_tuned = tuned_pipeline.predict(predict_input)
            preds_tuned = predicted_values_tuned.predict

            mae_value = mean_absolute_error(y_test, preds_tuned)

            print('Obtained metrics after tuning:')
            print(f'MAE - {mae_value:.4f}\n')

        pipeline.unfit()


# Script for testing is pipeline can process different datasets for regression task
if __name__ == '__main__':
    run_experiment(regression_ransac_pipeline(), tuner=SequentialTuner)



import os
from dataclasses import dataclass
from typing import List, Optional, Sequence, Any, TypeVar, Callable, Hashable

import numpy as np

from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.utilities.serializable import Serializable
from fedot.core.optimisers.gp_comp.evaluation import DelegateEvaluator
from fedot.remote.infrastructure.clients.client import Client
from fedot.utilities.pattern_wrappers import singleton


def init_data_for_remote_execution(train_data: InputData):
    setup = RemoteEvaluator()
    if setup.remote_task_params is not None:
        setup.remote_task_params.train_data_idx = train_data.idx


@dataclass
class RemoteTaskParams:
    """Class with parameters of remote evaluation.

    :param mode: evaluation mode - 'remote' or 'local'
    :param dataset_name: name of remote dataset used for fitting
    :param task_type: string representation of Task class for FEDOT
    :param train_data_idx: indices to subset dataset for fitting
    :param is_multi_modal: is train data multi-modal?
    :param var_names: variable names for features
    :param target: variable name for target
    :param max_parallel maximal number of parallel remote task
    """
    mode: str = 'local'
    dataset_name: Optional[str] = None
    task_type: Optional[str] = None
    train_data_idx: Optional[List] = None
    is_multi_modal: bool = False
    var_names: Optional[List] = None
    target: Optional[str] = None
    max_parallel: int = 7


G = TypeVar('G', bound=Serializable)


@singleton
class RemoteEvaluator(DelegateEvaluator):
    def __init__(self):
        """
        Class for the batch evaluation of pipelines using remote client
        """
        self._logger = default_log(prefix='RemoteFitterLog')
        self.remote_task_params = None
        self.client = None
        self.config_for_dump = _get_config

    def init(self, client: Client = None,
             remote_task_params: Optional[RemoteTaskParams] = None,
             get_config: Optional[Callable] = None):
        """
        :param client: client class for connection to external computational server.
        :param remote_task_params: dictionary with the parameters of remote evaluation.
        :param get_config: optional function that constructs config for remote client.

        """
        self.remote_task_params = remote_task_params
        self.client = client
        self.config_for_dump = get_config or _get_config

    @property
    def is_enabled(self):
        return self.remote_task_params is not None and self.remote_task_params.mode == 'remote'

    def compute_graphs(self, graphs: Sequence[G]) -> Sequence[G]:
        params = self.remote_task_params

        client = self.client
        execution_ids = {}
        graph_batches = _prepare_batches(graphs, params.max_parallel)
        final_graphs = []

        # start of the remote execution for each pipeline
        for graphs_batch in graph_batches:
            for graph in graphs_batch:
                task_id = self._create_graph_task(graph)
                execution_ids[id(graph)] = task_id

            # waiting for readiness of all pipelines
            ex_time = client.wait_until_ready()

            # download of remote execution result for each pipeline
            for p_id, graph in enumerate(graphs_batch):
                task_id = execution_ids.get(id(graph), None)
                if task_id:
                    try:
                        graphs_batch[p_id] = client.download_result(task_id)
                    except Exception as ex:
                        self._logger.warning(f'{p_id}, {ex}')
            final_graphs.extend(graphs_batch)

            self._logger.info(f'REMOTE EXECUTION TIME {ex_time}')

        return final_graphs

    def _create_graph_task(self, graph: G) -> Optional[Hashable]:
        """Serializes task and creates a graph task for remote client.

        :return: task id
        """

        graph_json, _ = graph.save()
        graph_json = graph_json.replace('\n', '')

        config = self.config_for_dump(graph_json, self.remote_task_params,
                                      self.client.exec_params, self.client.connect_params)

        task_id = self.client.create_task(config=config)
        return task_id


def _prepare_batches(graphs: Sequence[Any], max_parallel: int):
    num_parts = np.floor(len(graphs) / max_parallel)
    num_parts = max(num_parts, 1)
    pipelines_parts = [x.tolist() for x in np.array_split(graphs, num_parts)]
    return pipelines_parts


def _get_config(graph_json: dict, params: RemoteTaskParams, client_params: dict, conn_params: dict):
    var_names = list(map(str, params.var_names)) \
        if params.var_names is not None else []
    train_data_idx = list(map(str, params.train_data_idx)) \
        if params.train_data_idx is not None else []

    data_name = params.dataset_name
    if conn_params is not None and conn_params:
        train_data = f"{client_params['container_input_path']}/data/{conn_params['DATA_ID']}/{data_name}.csv"
    else:
        train_data = f"{client_params['container_input_path']}/{data_name}.csv"
    return f"""[DEFAULT]
        pipeline_template = {graph_json}
        train_data = {train_data}
        task = {params.task_type}
        output_path = {client_params['container_output_path']}
        train_data_idx = {train_data_idx}
        var_names = {var_names}
        is_multi_modal = {params.is_multi_modal}
        target = {params.target}
        [OPTIONAL]
        """.encode('utf-8')


def _init_from_env():
    return {'FEDOT_LOGIN': os.environ['FEDOT_LOGIN'],
            'FEDOT_PASSWORD': os.environ['FEDOT_PASSWORD'],
            'AUTH_SERVER': os.environ['AUTH_SERVER'],
            'CONTR_SERVER': os.environ['CONTR_SERVER'],
            'PROJECT_ID': os.environ['PROJECT_ID'],
            'DATA_ID': os.environ['DATA_ID']}



import os
from copy import deepcopy

import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from sklearn.metrics import mean_squared_error, mean_absolute_error

from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TsForecastingParams, TaskTypesEnum


def mean_absolute_percentage_error(y_true, y_pred):
    y_true, y_pred = np.array(y_true), np.array(y_pred)
    return round((np.mean(np.abs((y_true - y_pred) / y_true)) * 100), 4)


def prepare_data(time_series, exog_variable, len_forecast=250):
    time_series = np.array(time_series)
    exog_variable = np.array(exog_variable)

    train_input, predict_input = train_test_data_setup(
        InputData(idx=np.arange(len(time_series)),
                  features=time_series,
                  target=time_series,
                  task=Task(TaskTypesEnum.ts_forecasting,
                            TsForecastingParams(
                                forecast_length=len_forecast)),
                  data_type=DataTypesEnum.ts))

    # Exogenous time series
    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(
                    forecast_length=len_forecast))

    predict_input_exog = InputData(idx=np.arange(len(exog_variable)),
                                   features=exog_variable, target=time_series,
                                   task=task, data_type=DataTypesEnum.ts)
    train_input_exog, predict_input_exog = train_test_data_setup(predict_input_exog)

    return train_input, predict_input, train_input_exog, predict_input_exog, predict_input.target


def get_arima_nemo_pipeline():
    """ Function return complex pipeline with the following structure
        arima \
               linear
        nemo  |
    """

    node_arima = PrimaryNode('arima')
    node_nemo = PrimaryNode('exog_ts')
    node_final = SecondaryNode('linear', nodes_from=[node_arima, node_nemo])
    pipeline = Pipeline(node_final)
    return pipeline


def get_stlarima_nemo_pipeline():
    """ Function return complex pipeline with the following structure
        stl_arima \
                   linear
            nemo  |
    """

    node_arima = PrimaryNode('stl_arima')
    node_arima.parameters = {'period': 80, 'p': 2, 'd': 1, 'q': 0}
    node_nemo = PrimaryNode('exog_ts')
    node_final = SecondaryNode('linear', nodes_from=[node_arima, node_nemo])
    pipeline = Pipeline(node_final)
    return pipeline


def get_ridge_nemo_pipeline():
    """ Function return complex pipeline with the following structure
        lagged -> ridge \
                          ridge
        lagged -> ridge  |      \
                                 linear
                          nemo  /
    """

    node_lagged_1 = PrimaryNode('lagged/1')
    node_ridge_1 = SecondaryNode('ridge', nodes_from=[node_lagged_1])
    node_lagged_2 = PrimaryNode('lagged/2')
    node_ridge_2 = SecondaryNode('ridge', nodes_from=[node_lagged_2])
    node_ridge_3 = SecondaryNode('ridge', nodes_from=[node_ridge_1, node_ridge_2])
    node_nemo = PrimaryNode('exog_ts')
    node_final = SecondaryNode('linear', nodes_from=[node_ridge_3, node_nemo])
    pipeline = Pipeline(node_final)
    return pipeline


def get_arima_pipeline():
    """ Function return complex pipeline with the following structure
        arima
    """

    node_final = PrimaryNode('arima')

    pipeline = Pipeline(node_final)
    return pipeline


def get_stlarima_pipeline():
    """ Function return complex pipeline with the following structure
        stl_arima
    """

    node_final = PrimaryNode('stl_arima')
    node_final.parameters = {'period': 80, 'p': 2, 'd': 1, 'q': 0}
    pipeline = Pipeline(node_final)
    return pipeline


def get_ridge_pipeline():
    """ Function return complex pipeline with the following structure
        lagged -> ridge \
                          ridge
        lagged -> ridge  |
    """

    node_lagged_1 = PrimaryNode('lagged/1')
    node_ridge_1 = SecondaryNode('ridge', nodes_from=[node_lagged_1])

    node_lagged_2 = PrimaryNode('lagged/2')

    node_ridge_2 = SecondaryNode('ridge', nodes_from=[node_lagged_2])

    node_final = SecondaryNode('ridge', nodes_from=[node_ridge_1, node_ridge_2])
    pipeline = Pipeline(node_final)
    return pipeline


def compare_plot(predicted, real, forecast_length, model):
    plt.plot(real, linewidth=1, label="Observed", alpha=0.4)
    plt.plot(predicted, linewidth=1, label="Predicted", alpha=0.6)
    plt.legend()
    plt.xlabel('Time index')
    plt.ylabel('SSH, m')
    plt.title(f'Sea surface height forecast for {forecast_length} days with {model}')
    plt.show()


def run_nemo_based_forecasting(time_series, exog_variable, len_forecast=60, visualization=False):
    errors_df = {}

    train_input, predict_input, train_input_exog, predict_input_exog, test_data = \
        prepare_data(time_series=time_series,
                     exog_variable=exog_variable,
                     len_forecast=len_forecast)

    pipelines = {'ARIMA': {
        'tr_nodes_data': {"arima": train_input},
        'pr_nodes_data': {"arima": predict_input},
        'model': get_arima_pipeline()
    },
        'STL_ARIMA': {
            'tr_nodes_data': {"stl_arima": train_input},
            'pr_nodes_data': {"stl_arima": predict_input},
            'model': get_stlarima_pipeline()
        },
        'RIDGE':
            {'tr_nodes_data': {"lagged/1": train_input, "lagged/2": train_input},
             'pr_nodes_data': {"lagged/1": predict_input, "lagged/2": predict_input},
             'model': get_ridge_pipeline()
             },
        'ARIMA_NEMO':
            {'tr_nodes_data': {"arima": train_input, "exog_ts": train_input_exog},
             'pr_nodes_data': {"arima": predict_input,
                               "exog_ts": predict_input_exog},
             'model': get_arima_nemo_pipeline()
             },
        'STL_ARIMA_NEMO':
            {'tr_nodes_data': {"stl_arima": train_input,
                               "exog_ts": train_input_exog},
             'pr_nodes_data': {"stl_arima": predict_input,
                               "exog_ts": predict_input_exog},
             'model': get_stlarima_nemo_pipeline()
             },
        'RIDGE_NEMO':
            {'tr_nodes_data': {"lagged/1": train_input,
                               "lagged/2": train_input,
                               "exog_ts": train_input_exog},
             'pr_nodes_data': {"lagged/1": predict_input,
                               "lagged/2": predict_input,
                               "exog_ts": predict_input_exog},
             'model': get_ridge_nemo_pipeline()
             }
    }

    def multimodal_data_preparing(name):
        pipeline = pipelines[name]['model']
        train_dict = {}
        predict_dict = {}
        for key in pipelines[name]['tr_nodes_data'].keys():
            train_dict[key] = deepcopy(pipelines[name]['tr_nodes_data'][key])
        for key in pipelines[name]['pr_nodes_data'].keys():
            predict_dict[key] = deepcopy(pipelines[name]['pr_nodes_data'][key])

        train_dataset = MultiModalData(train_dict)
        predict_dataset = MultiModalData(predict_dict)
        return pipeline, train_dataset, predict_dataset

    for model_name in pipelines.keys():
        pipeline, train_dataset, predict_dataset = multimodal_data_preparing(model_name)

        pipeline.fit_from_scratch(train_dataset)
        predicted_values = pipeline.predict(predict_dataset)
        predicted_values = predicted_values.predict
        predicted = np.ravel(np.array(predicted_values))
        test_data = np.ravel(test_data)

        mse_before = mean_squared_error(test_data, predicted, squared=False)
        mae_before = mean_absolute_error(test_data, predicted)
        mape_before = mean_absolute_percentage_error(test_data, predicted)

        errors_df[model_name + '_MSE'] = mse_before
        errors_df[model_name + '_MAE'] = mae_before
        errors_df[model_name + '_MAPE'] = mape_before

        if visualization:
            compare_plot(predicted, test_data, len_forecast, model_name)
            print(model_name)
            print(f' MSE - {mse_before:.4f}')
            print(f' MAE - {mae_before:.4f}')
            print(f' MAPE - {mape_before:.4f}\n')

    return errors_df


def boxplot_visualize(df, label):
    ylabel = 'ssh, meters'
    if label == 'MAPE':
        ylabel = ''

    df.boxplot(rot=13)
    plt.ylabel(ylabel)
    plt.title(f'{label} distribution')

    plt.show()


def run_single_example(len_forecast=40, visualization=False):
    ts_name = 'sea_level'
    path_to_file = '../../cases/data/nemo/sea_surface_height.csv'
    path_to_exog_file = '../../cases/data/nemo/sea_surface_height_nemo.csv'

    df = pd.read_csv(path_to_file)
    time_series = df[ts_name]
    df = pd.read_csv(path_to_exog_file)
    exog_variable = df[ts_name]

    run_nemo_based_forecasting(time_series=time_series,
                               exog_variable=exog_variable,
                               len_forecast=len_forecast,
                               visualization=visualization)


def create_errors_df():
    df = pd.DataFrame(columns=['POINT', 'RIDGE', 'RIDGE_NEMO',
                               'ARIMA', 'ARIMA_NEMO',
                               'STL_ARIMA', 'STL_ARIMA_NEMO'])
    return df


def add_data_to_errors_df(df, error_name, point, errors):
    df = df.append({'POINT': point,
                    'RIDGE': errors['RIDGE_' + error_name],
                    'RIDGE_NEMO': errors['RIDGE_NEMO_' + error_name],
                    'ARIMA': errors['ARIMA_' + error_name],
                    'ARIMA_NEMO': errors['ARIMA_NEMO_' + error_name],
                    'STL_ARIMA': errors['STL_ARIMA_' + error_name],
                    'STL_ARIMA_NEMO': errors['STL_ARIMA_NEMO_' + error_name],
                    }, ignore_index=True)
    return df


def run_multiple_example(path_to_file, path_to_exog_file, out_path=None, visualization=False, len_forecast=40):
    mse_errors_df = create_errors_df()
    mae_errors_df = create_errors_df()
    mape_errors_df = create_errors_df()

    path_to_file = path_to_file
    path_to_exog_file = path_to_exog_file

    df = pd.read_csv(path_to_file)
    df_exog = pd.read_csv(path_to_exog_file)

    for point in df.columns.values:
        if point != 'dates':
            time_series = df[point]
            exog_variable = df_exog[point]

            errors = run_nemo_based_forecasting(time_series=time_series,
                                                exog_variable=exog_variable,
                                                len_forecast=len_forecast,
                                                visualization=visualization)

            mse_errors_df = add_data_to_errors_df(mse_errors_df, 'MSE', point, errors)
            mae_errors_df = add_data_to_errors_df(mae_errors_df, 'MAE', point, errors)
            mape_errors_df = add_data_to_errors_df(mape_errors_df, 'MAPE', point, errors)

    if out_path is not None:
        mse_errors_df.to_csv(os.path.join(out_path, 'mse_errors.csv'), index=False)
        mae_errors_df.to_csv(os.path.join(out_path, 'mae_errors.csv'), index=False)
        mape_errors_df.to_csv(os.path.join(out_path, 'mape_errors.csv'), index=False)

    if visualization:
        boxplot_visualize(mse_errors_df, 'MSE')
        boxplot_visualize(mae_errors_df, 'MAE')
        boxplot_visualize(mape_errors_df, 'MAPE')


def run_prediction_examples(mode='single', visualization=False):
    if mode == 'single':
        run_single_example(len_forecast=40, visualization=visualization)
    if mode == 'multiple':
        run_multiple_example(path_to_file='../../cases/data/nemo/SSH_points_grid.csv',
                             path_to_exog_file='../../cases/data/nemo/SSH_nemo_points_grid.csv',
                             out_path=None,
                             len_forecast=30,
                             visualization=visualization)


if __name__ == '__main__':
    run_prediction_examples(mode='multiple', visualization=True)



from sklearn.metrics import roc_auc_score as roc_auc

from cases.data.data_utils import get_scoring_case_data_paths
from fedot.core.data.data import InputData
from fedot.core.pipelines.node import PrimaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.utilities.project_import_export import export_project_to_zip, import_project_from_zip

if __name__ == '__main__':
    train_file_path, test_file_path = get_scoring_case_data_paths()
    train_data = InputData.from_csv(train_file_path)
    test_data = InputData.from_csv(test_file_path)

    pipeline = Pipeline(PrimaryNode('rf'))

    # Export project to zipfile to directory
    export_project_to_zip(zip_name='example', opt_history=None,
                          pipeline=pipeline, train_data=train_data, test_data=test_data)

    # Import project from zipfile to pipeline and InputData objects.
    pipeline, train_data, test_data, opt_history = \
        import_project_from_zip('example')

    pipeline.fit(train_data)
    prediction = pipeline.predict(test_data)

    print(roc_auc(test_data.target, prediction.predict))



from fedot.core.log import default_log
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.pipelines.pipeline_builder import PipelineBuilder


def generate_pipeline() -> Pipeline:
    node_scaling = PrimaryNode('scaling')
    node_first = SecondaryNode('kmeans', nodes_from=[node_scaling])
    node_second = SecondaryNode('rf', nodes_from=[node_scaling])
    node_third = SecondaryNode('linear', nodes_from=[node_scaling])
    node_root = SecondaryNode('logit', nodes_from=[node_first, node_second, node_third, node_scaling])

    return Pipeline(node_root)


def show_default(pipeline: Pipeline):
    """ Show with default properties via networkx. """
    pipeline.show()


def show_customized(pipeline: Pipeline):
    """ Show with adjusted sizes and green nodes. """
    pipeline.show(node_color='green', edge_curvature_scale=1.2, node_size_scale=2.0, font_size_scale=1.4)


def show_custom_colors(pipeline: Pipeline):
    """ Show with colors defined by label-color dictionary. """
    pipeline.show(node_color={'scaling': 'tab:olive', 'rf': '#FF7F50', 'linear': (0, 1, 1), None: 'black'},
                  edge_curvature_scale=1.2, node_size_scale=2.0, font_size_scale=1.4)


def show_complex_colors(pipeline: Pipeline):
    """ Show with colors defined by function. """
    def nodes_color(labels):
        if 'xgboost' in labels:
            return {'xgboost': 'tab:orange', None: 'black'}
        else:
            return {'rf': 'tab:green', None: 'black'}

    pipeline.show(node_color=nodes_color)


def show_pyvis(pipeline: Pipeline):
    """ Show with pyvis. """
    pipeline.show(engine='pyvis')


def show_pyvis_custom_colors(pipeline: Pipeline):
    """ Show with pyvis with custom colors. """
    pipeline.show(engine='pyvis',
                  node_color={'scaling': 'tab:olive', 'rf': '#FF7F50', 'linear': (0, 1, 1), None: 'black'})


def show_graphviz(pipeline: Pipeline):
    """ Show with graphviz (requires Graphviz and pygraphviz). """
    pipeline.show(engine='graphviz')


def show_graphviz_custom_colors(pipeline: Pipeline):
    """ Show with graphviz with custom colors (requires Graphviz and pygraphviz). """
    pipeline.show(engine='graphviz',
                  node_color={'scaling': 'tab:olive', 'rf': '#FF7F50', 'linear': (0, 1, 1), None: 'black'})


def main():
    pipeline = generate_pipeline()
    show_default(pipeline)
    show_customized(pipeline)
    show_custom_colors(pipeline)
    show_complex_colors(pipeline)
    show_complex_colors(PipelineBuilder(*pipeline.nodes).add_node('xgboost').to_pipeline())
    show_pyvis(pipeline)
    show_pyvis_custom_colors(pipeline)
    try:
        import graphviz
        show_graphviz(pipeline)
        show_graphviz_custom_colors(pipeline)
    except ImportError:
        default_log().info('Either Graphviz or pygraphviz is not installed. Skipping visualizations.')


if __name__ == '__main__':
    main()



import json
import os

import numpy as np

from examples.simple.regression.regression_with_tuning import get_regression_dataset
from examples.simple.regression.regression_pipelines import regression_ransac_pipeline
from fedot.core.data.data import InputData
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import Task, TaskTypesEnum


def create_correct_path(path: str, dirname_flag: bool = False):
    """
    Create path with time which was created during the testing process.
    """
    last_el = None
    for dirname in next(os.walk(os.path.curdir))[1]:
        if dirname.endswith(path):
            if dirname_flag:

                last_el = dirname
            else:
                file = os.path.join(dirname, path + '.json')
                last_el = file
    return last_el


def run_import_export_example(pipeline_path, pipeline):
    features_options = {'informative': 1, 'bias': 0.0}
    samples_amount = 100
    features_amount = 2
    x_train, y_train, x_test, y_test = get_regression_dataset(features_options,
                                                              samples_amount,
                                                              features_amount)

    # Define regression task
    task = Task(TaskTypesEnum.regression)

    # Prepare data to train the model
    train_input = InputData(idx=np.arange(0, len(x_train)),
                            features=x_train,
                            target=y_train,
                            task=task,
                            data_type=DataTypesEnum.table)

    predict_input = InputData(idx=np.arange(0, len(x_test)),
                              features=x_test,
                              target=None,
                              task=task,
                              data_type=DataTypesEnum.table)

    # Get pipeline and fit it
    pipeline.fit_from_scratch(train_input)

    predicted_output = pipeline.predict(predict_input)
    prediction_before_export = np.array(predicted_output.predict)
    print(f'Before export {prediction_before_export[:4]}')

    # Export it
    pipeline.save(path=pipeline_path)

    # Import pipeline
    json_path_load = create_correct_path(pipeline_path)
    new_pipeline = Pipeline.from_serialized(json_path_load)

    predicted_output_after_export = new_pipeline.predict(predict_input)
    prediction_after_export = np.array(predicted_output_after_export.predict)

    print(f'After import {prediction_after_export[:4]}')

    dict_pipeline, dict_fitted_operations = pipeline.save()
    dict_pipeline = json.loads(dict_pipeline)
    pipeline_from_dict = Pipeline.from_serialized(dict_pipeline, dict_fitted_operations)

    predicted_output = pipeline_from_dict.predict(predict_input)
    prediction = np.array(predicted_output.predict)
    print(f'Prediction from pipeline loaded from dict {prediction[:4]}')


if __name__ == '__main__':
    run_import_export_example(pipeline_path='import_export', pipeline=regression_ransac_pipeline())



from datetime import timedelta
from typing import Callable, Type, Union

from hyperopt import tpe

from fedot.core.data.data import InputData
from fedot.core.optimisers.composer_requirements import ComposerRequirements
from fedot.core.optimisers.objective import PipelineObjectiveEvaluate
from fedot.core.optimisers.objective.data_source_splitter import DataSourceSplitter
from fedot.core.optimisers.objective.metrics_objective import MetricsObjective
from fedot.core.pipelines.tuning.search_space import SearchSpace
from fedot.core.pipelines.tuning.tuner_interface import HyperoptTuner
from fedot.core.pipelines.tuning.unified import PipelineTuner
from fedot.core.repository.quality_metrics_repository import MetricType, MetricsEnum
from fedot.core.repository.tasks import Task
from fedot.utilities.define_metric_by_task import MetricByTask


class TunerBuilder:
    def __init__(self, task: Task):
        self.tuner_class = PipelineTuner
        self.cv_folds = None
        self.validation_blocks = None
        self.n_jobs = -1
        self.metric: MetricsEnum = MetricByTask.get_default_quality_metrics(task.task_type)[0]
        self.iterations = 100
        self.early_stopping_rounds = None
        self.timeout = timedelta(minutes=5)
        self.search_space = SearchSpace()
        self.algo = tpe.suggest
        self.eval_time_constraint = None

    def with_tuner(self, tuner: Type[HyperoptTuner]):
        self.tuner_class = tuner
        return self

    def with_requirements(self, requirements: ComposerRequirements):
        self.cv_folds = requirements.cv_folds
        self.validation_blocks = requirements.validation_blocks
        self.n_jobs = requirements.n_jobs
        return self

    def with_cv_folds(self, cv_folds: int):
        self.cv_folds = cv_folds
        return self

    def with_validation_blocks(self, validation_blocks: int):
        self.validation_blocks = validation_blocks
        return self

    def with_n_jobs(self, n_jobs: int):
        self.n_jobs = n_jobs
        return self

    def with_metric(self, metric: MetricType):
        self.metric = metric
        return self

    def with_iterations(self, iterations: int):
        self.iterations = iterations
        return self

    def with_early_stopping_rounds(self, early_stopping_rounds: int):
        self.early_stopping_rounds = early_stopping_rounds
        return self

    def with_timeout(self, timeout: timedelta):
        self.timeout = timeout
        return self

    def with_eval_time_constraint(self, eval_time_constraint: Union[timedelta, int, float]):
        self.eval_time_constraint = eval_time_constraint
        return self

    def with_search_space(self, search_space: SearchSpace):
        self.search_space = search_space
        return self

    def with_algo(self, algo: Callable):
        self.algo = algo
        return self

    def build(self, data: InputData) -> HyperoptTuner:
        objective = MetricsObjective(self.metric)
        data_producer = DataSourceSplitter(self.cv_folds, self.validation_blocks).build(data)
        objective_evaluate = PipelineObjectiveEvaluate(objective, data_producer,
                                                       validation_blocks=self.validation_blocks,
                                                       time_constraint=self.eval_time_constraint)
        tuner = self.tuner_class(objective_evaluate=objective_evaluate,
                                 iterations=self.iterations,
                                 early_stopping_rounds=self.early_stopping_rounds,
                                 timeout=self.timeout,
                                 search_space=self.search_space,
                                 algo=self.algo,
                                 n_jobs=self.n_jobs)
        return tuner



import json
from os.path import join
from typing import List, Optional, Type

from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.utils import default_fedot_data_dir
from fedot.sensitivity.operations_hp_sensitivity.multi_operations_sensitivity import MultiOperationsHPAnalyze
from fedot.sensitivity.sa_requirements import SensitivityAnalysisRequirements


class PipelineAnalysis:
    """This class is for analyzing the :class:`Pipeline` as the black-box model,
    using analysis approaches defined for whole pipeline perturbation, i.e. :class:`MultiOperationsHPAnalyze`

    Args:
        pipeline: pipeline object to analyze
        train_data: data used for Pipeline training
        test_data: data used for Pipeline validation
        approaches: methods applied to pipeline Default: :class:`MultiOperationsHPAnalyze`
        requirements: extra requirements to define specific details for different approaches
            See :class:`SensitivityAnalysisRequirements` class documentation
        path_to_save: path to save results to. Default: ``~home/Fedot/sensitivity/pipeline_sa``
    """

    def __init__(self, pipeline: Pipeline, train_data: InputData, test_data: InputData,
                 approaches: Optional[List[Type[MultiOperationsHPAnalyze]]] = None,
                 requirements: SensitivityAnalysisRequirements = None,
                 path_to_save=None):
        self.pipeline = pipeline
        self.train_data = train_data
        self.test_data = test_data
        self.requirements = \
            SensitivityAnalysisRequirements() if requirements is None else requirements
        self.approaches = [MultiOperationsHPAnalyze] if approaches is None else approaches
        self.path_to_save = \
            join(default_fedot_data_dir(), 'sensitivity', 'pipeline_sa') if path_to_save is None else path_to_save

        self.log = default_log(self)

    def analyze(self) -> dict:
        """Apply defined approaches for the black-box pipeline analysis
        """

        all_approaches_results = dict()
        for approach in self.approaches:
            analyze_result = approach(pipeline=self.pipeline,
                                      train_data=self.train_data,
                                      test_data=self.test_data,
                                      requirements=self.requirements).analyze()
            all_approaches_results[f'{approach.__name__}'] = analyze_result

        if self.requirements.is_save:
            self._save_results_to_json(all_approaches_results)

        return all_approaches_results

    def _save_results_to_json(self, result: dict):
        result_file = join(self.path_to_save, 'pipeline_SA_results.json')
        with open(result_file, 'w', encoding='utf-8') as file:
            file.write(json.dumps(result, indent=4))



import numpy as np
from typing import Dict, Tuple, Any

from fedot.core.data.data import InputData
from fedot.core.data.data_preprocessing import find_categorical_columns
from fedot.core.data.multi_modal import MultiModalData
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.preprocessing.structure import DEFAULT_SOURCE_NAME


class DataAnalyser:
    """
    Class to analyse data that comes to FEDOT API.
    All methods are inplace to prevent unnecessary copy of large datasets
    It functionality is:
    1) Cut large datasets to prevent memory stackoverflow
    2) Use label encoder with tree models instead OneHot when summary cardinality of categorical features is high
    """

    def __init__(self, safe_mode: bool):
        self.safe_mode = safe_mode
        self.max_size = 50000000
        self.max_cat_cardinality = 50

    # TODO implement correct logic to process multimodal data
    def give_recommendation(self, input_data: InputData, source_name: str = DEFAULT_SOURCE_NAME) -> Dict:
        """
        Gives a recommendation of cutting dataset or using label encoding
        :param input_data: data for preprocessing
        :param source_name: name of data source node
        :return : dict with str recommendations
        """

        recommendations = {}
        if isinstance(input_data, MultiModalData):
            for data_source_name, values in input_data.items():
                recommendations[data_source_name] = self.give_recommendation(input_data[data_source_name],
                                                                             data_source_name)
        elif isinstance(input_data, InputData) and input_data.data_type == DataTypesEnum.table:
            if self.safe_mode:
                is_cut_needed, border = self.control_size(input_data)
                if is_cut_needed:
                    recommendations['cut'] = {'border': border}
                is_label_encoding_needed = self.control_categorical(input_data)
                if is_label_encoding_needed:
                    recommendations['label_encoded'] = {}
        return recommendations

    def control_size(self, input_data: InputData) -> Tuple[bool, Any]:
        """
        Check if size of table (N*M) > threshold and cutting is needed
        :param input_data: data for preprocessing

        :return : (is_cut_needed, border) is cutting is needed | if yes - border of cutting,
        """

        if input_data.data_type == DataTypesEnum.table:
            if input_data.features.shape[0] * input_data.features.shape[1] > self.max_size:
                border = self.max_size // input_data.features.shape[1]
                return True, border
        return False, None

    def control_categorical(self, input_data: InputData) -> bool:
        """
        Check if use label encoder instead oneHot if summary cardinality > threshold

        :param input_data: data for preprocessing
        """

        categorical_ids, _ = find_categorical_columns(input_data.features)
        all_cardinality = 0
        need_label = False
        for idx in categorical_ids:
            all_cardinality += np.unique(input_data.features[:, idx].astype(str)).shape[0]
            if all_cardinality > self.max_cat_cardinality:
                need_label = True
                break
        return need_label



from copy import copy
from random import randint
from typing import (Any, Callable, List, Sequence)

from numpy import random

from fedot.core.composer.composer import Composer
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.optimisers.fitness import Fitness
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.objective import Objective, ObjectiveFunction
from fedot.core.optimisers.optimizer import GraphOptimizer
from fedot.core.pipelines.node import SecondaryNode, PrimaryNode
from fedot.core.pipelines.pipeline import Node, Pipeline


class RandomSearchComposer(Composer):
    def __init__(self, optimizer: 'RandomSearchOptimizer'):
        super().__init__(optimizer=optimizer)

    def compose_pipeline(self, data: InputData) -> Pipeline:
        train_data, test_data = train_test_data_setup(data, shuffle_flag=True)

        def prepared_objective(pipeline: Pipeline) -> Fitness:
            pipeline.fit(train_data)
            return self.optimizer.objective(pipeline, reference_data=test_data)

        best_pipeline = self.optimizer.optimise(prepared_objective)[0]
        return best_pipeline


def nodes_to_pipeline(nodes: List[Node]) -> Pipeline:
    pipeline = Pipeline()
    [pipeline.add_node(nodes) for nodes in nodes]
    return pipeline


class RandomGraphFactory:
    def __init__(self,
                 primary_candidates: Sequence[Any], secondary_candidates: Sequence[Any],
                 primary_node_func: Callable = PrimaryNode, secondary_node_func: Callable = SecondaryNode):
        self.__primary_node_func = primary_node_func
        self.__secondary_node_func = secondary_node_func
        self.__primary_candidates = list(primary_candidates)
        self.__secondary_candidates = list(secondary_candidates)

    def __call__(self):
        return self.random_nodeset()

    def random_nodeset(self) -> Pipeline:
        new_set = []

        # random primary nodes
        num_of_primary = randint(1, len(self.__primary_candidates))
        for _ in range(num_of_primary):
            new_set.append(self.random_primary())

        # random final node
        if len(new_set) > 1:
            parent_nodes = copy(new_set)
            final_node = self.random_secondary(parent_nodes)
            new_set.append(final_node)

        return nodes_to_pipeline(new_set)

    def random_primary(self):
        return self.__primary_node_func(random.choice(self.__primary_candidates))

    def random_secondary(self, parent_nodes):
        return self.__secondary_node_func(random.choice(self.__secondary_candidates), parent_nodes)


class RandomSearchOptimizer(GraphOptimizer):

    def __init__(self, objective: Objective,
                 requirements: PipelineComposerRequirements,
                 ):
        super().__init__(objective, requirements=requirements)
        self._factory = RandomGraphFactory(requirements.primary, requirements.secondary)

    def optimise(self, objective: ObjectiveFunction) -> Sequence[Pipeline]:
        best_metric_value = 1000
        best_set = None
        history = []
        for i in range(self.requirements.num_of_generations):
            new_pipeline = self._factory()
            new_metric_value = objective(new_pipeline).value
            new_metric_value = round(new_metric_value, 3)
            if new_metric_value < best_metric_value:
                best_metric_value = new_metric_value
                best_set = new_pipeline

            history.append((new_pipeline, new_metric_value))
            self.log.info(f'Iter {i}: best metric {best_metric_value},'
                          f'try {new_metric_value} with num nodes {new_pipeline.length}')
        return [best_set]



import traceback
from datetime import timedelta
from typing import Callable, Iterable, Optional, Tuple

import numpy as np

from fedot.core.caching.pipelines_cache import OperationsCache
from fedot.core.caching.preprocessing_cache import PreprocessingCache
from fedot.core.data.data import InputData
from fedot.core.log import default_log
from fedot.core.operations.model import Model
from fedot.core.optimisers.fitness import Fitness
from fedot.core.pipelines.pipeline import Pipeline
from fedot.utilities.debug import is_recording_mode, is_test_session, save_debug_info_for_pipeline
from .objective import Objective, to_fitness
from .objective_eval import ObjectiveEvaluate

DataSource = Callable[[], Iterable[Tuple[InputData, InputData]]]


class PipelineObjectiveEvaluate(ObjectiveEvaluate[Pipeline]):
    """
    Evaluator of Objective that requires train and test data for metric evaluation.
    Its role is to prepare graph on train-data and then evaluate metrics on test data.

    :param objective: Objective for evaluating metrics on pipelines.
    :param data_producer: Producer of data folds, each fold is a tuple of (train_data, test_data).
    If it returns a single fold, it's effectively a hold-out validation. For many folds it's k-folds.
    :param time_constraint: Optional time constraint for pipeline.fit.
    :param validation_blocks: Number of validation blocks, optional, used only for time series validation.
    :param pipelines_cache: Cache manager for fitted models, optional.
    :param preprocessing_cache: Cache manager for optional preprocessing encoders and imputers, optional.
    :param eval_n_jobs: number of jobs used to evaluate the objective.
    """

    def __init__(self,
                 objective: Objective,
                 data_producer: DataSource,
                 time_constraint: Optional[timedelta] = None,
                 validation_blocks: Optional[int] = None,
                 pipelines_cache: Optional[OperationsCache] = None,
                 preprocessing_cache: Optional[PreprocessingCache] = None,
                 eval_n_jobs: int = 1):
        super().__init__(objective, eval_n_jobs=eval_n_jobs)
        self._data_producer = data_producer
        self._time_constraint = time_constraint
        self._validation_blocks = validation_blocks
        self._pipelines_cache = pipelines_cache
        self._preprocessing_cache = preprocessing_cache
        self._log = default_log(self)

    def evaluate(self, graph: Pipeline) -> Fitness:
        # Seems like a workaround for situation when logger is lost
        #  when adapting and restoring it to/from OptGraph.
        graph.log = self._log

        graph_id = graph.root_node.descriptive_id
        self._log.debug(f'Pipeline {graph_id} fit started')

        folds_metrics = []
        for fold_id, (train_data, test_data) in enumerate(self._data_producer()):
            try:
                prepared_pipeline = self.prepare_graph(graph, train_data, fold_id, self._eval_n_jobs)
            except Exception as ex:
                self._log.warning(f'Error on Pipeline fit during fitness evaluation. '
                                  f'Skipping the Pipeline. Error <{ex}> on graph: {graph_id}')
                if is_test_session() and not isinstance(ex, TimeoutError):
                    stack_trace = traceback.format_exc()
                    save_debug_info_for_pipeline(graph, train_data, test_data, ex, stack_trace)
                    if not is_recording_mode():
                        raise ex
                continue

            evaluated_fitness = self._objective(prepared_pipeline,
                                                reference_data=test_data,
                                                validation_blocks=self._validation_blocks)
            if evaluated_fitness.valid:
                folds_metrics.append(evaluated_fitness.values)
            else:
                self._log.warning(f'Invalid fitness after objective evaluation. '
                                  f'Skipping the graph: {graph_id}')
                if is_test_session():
                    raise ValueError(f'Fitness {evaluated_fitness} is not valid')
                else:
                    continue
            graph.unfit()
        if folds_metrics:
            folds_metrics = tuple(np.mean(folds_metrics, axis=0))  # averages for each metric over folds
            self._log.debug(f'Pipeline {graph_id} with evaluated metrics: {folds_metrics}')
        else:
            folds_metrics = None

        return to_fitness(folds_metrics, self._objective.is_multi_objective)

    def prepare_graph(self, graph: Pipeline, train_data: InputData,
                      fold_id: Optional[int] = None, n_jobs: int = -1) -> Pipeline:
        """
        Fit pipeline before metric evaluation can be performed.
        :param graph: pipeline for train & validation
        :param train_data: InputData for training pipeline
        :param fold_id: id of the fold in cross-validation, used for cache requests.
        :param n_jobs: number of parallel jobs for preparation
        """
        if graph.is_fitted:
            # the expected behaviour for the remote evaluation
            return graph

        graph.unfit()

        # load preprocessing
        graph.try_load_from_cache(self._pipelines_cache, self._preprocessing_cache, fold_id)
        graph.fit(
            train_data,
            n_jobs=n_jobs,
            time_constraint=self._time_constraint
        )

        if self._pipelines_cache is not None:
            self._pipelines_cache.save_pipeline(graph, fold_id)
        if self._preprocessing_cache is not None:
            self._preprocessing_cache.add_preprocessor(graph, fold_id)

        return graph

    def evaluate_intermediate_metrics(self, graph: Pipeline):
        """Evaluate intermediate metrics"""
        # Get the last fold
        last_fold = None
        fold_id = None
        for fold_id, last_fold in enumerate(self._data_producer()):
            pass
        # And so test only on the last fold
        train_data, test_data = last_fold
        graph.try_load_from_cache(self._pipelines_cache, self._preprocessing_cache, fold_id)
        for node in graph.nodes:
            if not isinstance(node.operation, Model):
                continue
            intermediate_graph = Pipeline(node)
            intermediate_graph.fit(
                train_data,
                time_constraint=self._time_constraint,
                n_jobs=self._eval_n_jobs,
            )
            intermediate_fitness = self._objective(intermediate_graph,
                                                   reference_data=test_data,
                                                   validation_blocks=self._validation_blocks)
            # saving only the most important first metric
            node.metadata.metric = intermediate_fitness.values[0]



from typing import Callable, Optional

import numpy as np
from sklearn.neighbors import KNeighborsClassifier, KNeighborsRegressor

from fedot.core.operations.evaluation.operation_implementations.implementation_interfaces import ModelImplementation
from fedot.core.operations.operation_parameters import OperationParameters


class KNeighborsImplementation(ModelImplementation):

    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        self.model = None

    def fit(self, train_data):
        """ Method fit model on a dataset

        :param train_data: data to train the model
        """

        raise NotImplementedError()

    def predict(self, input_data):
        """ Method for making prediction

        :param input_data: data with features to process
        """
        prediction = self.model.predict(input_data.features)

        return prediction

    def check_and_correct_k_value(self, input_data, model_impl: Callable):
        """ Method check if the amount of neighbors is too big - clip it

        :param input_data: InputData for fit
        :param model_impl: Model to use
        """
        current_params = self.model.get_params()
        n_neighbors = current_params.get('n_neighbors')

        if n_neighbors > len(input_data.features):
            # Improve the parameter "n_neighbors": n_neighbors <= n_samples
            new_k_value = round(len(input_data.features) / 2)
            if new_k_value == 0:
                new_k_value = 1
            self.params.update(n_neighbors=new_k_value)
            self.model = model_impl(**self.params.to_dict())

            prefix = "n_neighbors of K-nn model was changed"
            self.log.info(f"{prefix} from {n_neighbors} to {new_k_value}")


class FedotKnnClassImplementation(KNeighborsImplementation):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        params = round_n_neighbors(self.params.to_dict())
        self.model = KNeighborsClassifier(**params)
        self.classes = None

    def fit(self, train_data):
        """ Method fit model on a dataset

        :param train_data: data to train the model
        """
        self.classes = np.unique(np.array(train_data.target))

        # Improve hyperparameters for model
        self.check_and_correct_k_value(train_data, KNeighborsClassifier)
        self.model.fit(train_data.features, train_data.target)
        return self.model

    def predict_proba(self, input_data):
        """ Method make prediction with probabilities of classes

        :param input_data: data with features to process
        """
        prediction = self.model.predict_proba(input_data.features)

        return prediction

    @property
    def classes_(self):
        return self.classes


class FedotKnnRegImplementation(KNeighborsImplementation):
    def __init__(self, params: Optional[OperationParameters] = None):
        super().__init__(params)
        params = round_n_neighbors(self.params.to_dict())
        self.model = KNeighborsRegressor(**params)

    def fit(self, train_data):
        """ Method fit model on a dataset

        :param train_data: data to train the model
        """

        # Improve hyperparameters for model
        self.check_and_correct_k_value(train_data, KNeighborsRegressor)
        self.model.fit(train_data.features, train_data.target)
        return self.model


def round_n_neighbors(params: dict) -> dict:
    """ Convert n_neighbors into integer value. Operation work inplace. """
    if 'n_neighbors' in params:
        n_neighbors = round(params['n_neighbors'])
        if n_neighbors == 0:
            n_neighbors = 1
        params['n_neighbors'] = n_neighbors
    return params



import datetime
from contextlib import contextmanager
from typing import Optional

from fedot.core.constants import COMPOSING_TUNING_PROPORTION, MINIMAL_PIPELINE_NUMBER_FOR_EVALUATION, \
    MINIMAL_SECONDS_FOR_TUNING


class ApiTime:
    """
    A class for performing operations on the AutoML algorithm runtime at the API level

    :param time_for_automl: time for AutoML algorithms in minutes
    """

    def __init__(self, **time_params):
        self.time_for_automl = time_params.get('time_for_automl')
        self.with_tuning = time_params.get('with_tuning')

        self.__define_timeouts_for_stages()

        self.composing_spend_time = datetime.timedelta(minutes=0)

        self.tuning_spend_time = datetime.timedelta(minutes=0)

        self.assumption_fit_spend_time = datetime.timedelta(minutes=0)

    def __define_timeouts_for_stages(self):
        """ Determine timeouts for tuning and composing """
        if self.time_for_automl in [None, -1]:
            self.timeout_for_composing = None
        else:
            # Time for composing based on tuning parameters
            if self.with_tuning:
                self.timeout_for_composing = self.time_for_automl * COMPOSING_TUNING_PROPORTION
            else:
                self.timeout_for_composing = self.time_for_automl

    def have_time_for_composing(self, pop_size: int, n_jobs: int) -> bool:
        timeout_not_set = self.timedelta_composing is None
        return timeout_not_set or self.assumption_fit_spend_time < self.timedelta_composing * n_jobs / pop_size

    def have_time_for_the_best_quality(self, n_jobs: int):
        timeout_not_set = self.timedelta_automl is None
        if timeout_not_set:
            return True
        return self.assumption_fit_spend_time <= self.timedelta_automl * n_jobs / MINIMAL_PIPELINE_NUMBER_FOR_EVALUATION

    def have_time_for_tuning(self):
        timeout_for_tuning = self.determine_resources_for_tuning()
        return timeout_for_tuning >= MINIMAL_SECONDS_FOR_TUNING

    @contextmanager
    def launch_composing(self):
        """ Wrap composing process with timer """
        starting_time_for_composing = datetime.datetime.now()
        yield
        self.composing_spend_time = datetime.datetime.now() - starting_time_for_composing

    @contextmanager
    def launch_tuning(self):
        """ Wrap tuning process with timer """
        starting_time_for_tuning = datetime.datetime.now()
        yield
        self.tuning_spend_time = datetime.datetime.now() - starting_time_for_tuning

    @contextmanager
    def launch_assumption_fit(self):
        """ Wrap assumption fit process with timer """
        starting_time_for_assumption_fit = datetime.datetime.now()
        yield
        self.assumption_fit_spend_time = datetime.datetime.now() - starting_time_for_assumption_fit

    def determine_resources_for_tuning(self):
        """
        Based on time spend for composing and initial pipeline fit determine
        how much time and how many iterations are needed for tuning

        """
        all_spend_time = self.composing_spend_time + self.assumption_fit_spend_time

        if self.time_for_automl is not None:
            all_timeout = float(self.time_for_automl)
            timeout_in_sec = datetime.timedelta(minutes=all_timeout).total_seconds()
            timeout_for_tuning = timeout_in_sec - all_spend_time.total_seconds()
        else:
            timeout_for_tuning = all_spend_time.total_seconds()
        return timeout_for_tuning

    @property
    def timedelta_composing(self) -> Optional[datetime.timedelta]:
        if self.timeout_for_composing is None:
            return None
        return datetime.timedelta(minutes=self.timeout_for_composing)

    @property
    def timedelta_automl(self) -> Optional[datetime.timedelta]:
        if self.time_for_automl is None:
            return None
        return datetime.timedelta(minutes=self.time_for_automl)



import logging
import operator
import timeit
from functools import reduce
from typing import Optional

import pandas as pd
from matplotlib import cm, colors, pyplot as plt

from fedot.api.main import Fedot
from fedot.core.optimisers.opt_history_objects.opt_history import OptHistory
from fedot.core.utils import fedot_project_root


def _count_pipelines(opt_history: Optional[OptHistory]) -> int:
    if opt_history is not None:
        return reduce(operator.add, map(len, opt_history.individuals), 0)
    return 0


def run_experiments(timeout: float = None, partitions_n=10, n_jobs=-1):
    """
    Performs experiment to show how much better to use multiprocessing mode in FEDOT

    :param timeout: timeout for optimization in minutes
    :param partitions_n: on how many folds you want. f.e. if dataset contains 20000 rows, partitions_n=5 will create
        such folds: [4000 rows, 8000 rows, 12000 rows, 16000 rows, 20000 rows]
    :param n_jobs: how many processors you want to use in a multiprocessing mode

    """
    train_data_path = f'{fedot_project_root()}/cases/data/scoring/scoring_train.csv'

    problem = 'classification'

    train_data = pd.read_csv(train_data_path)

    data_len = len(train_data)

    partitions = [int(data_len * (i / partitions_n)) for i in range(1, partitions_n + 1)]

    pipelines_count, times = [{1: [], n_jobs: []} for _ in range(2)]

    for _n_jobs in [1, n_jobs]:
        print(f'n_jobs: {_n_jobs}')
        for partition in partitions:
            train_data_tmp = train_data.iloc[:partition].copy()
            start_time = timeit.default_timer()
            auto_model = Fedot(problem=problem, seed=42, timeout=timeout,
                               n_jobs=_n_jobs, logging_level=logging.NOTSET,
                               with_tuning=False, preset='fast_train')
            auto_model.fit(features=train_data_tmp, target='target')
            times[_n_jobs].append((timeit.default_timer() - start_time) / 60)
            c_pipelines = _count_pipelines(auto_model.history)
            pipelines_count[_n_jobs].append(c_pipelines)
            print(f'\tDataset length: {partition}, number of pipelines: {c_pipelines}')

    plt.title('Comparison parallel mode with a single mode')
    plt.xlabel('rows in train dataset')
    plt.ylabel('Num of pipelines that were evaluated correctly')
    c_norm = colors.Normalize(vmin=timeout - timeout / 2, vmax=timeout + timeout / 2)
    plt.plot(partitions, pipelines_count[1], label='one process', zorder=1)
    plt.scatter(partitions, pipelines_count[1], c=times[1],
                cmap=cm.get_cmap('cool'), norm=c_norm, zorder=2)
    plt.plot(partitions, pipelines_count[n_jobs], label=f'{n_jobs} processes', zorder=1)
    plt.scatter(partitions, pipelines_count[n_jobs], c=times[n_jobs],
                cmap=cm.get_cmap('cool'), norm=c_norm, zorder=2)
    print(times)
    cb = plt.colorbar(cm.ScalarMappable(norm=c_norm, cmap=cm.get_cmap('cool')))
    cb.ax.set_ylabel('time for optimization in minutes', rotation=90)
    plt.legend()
    plt.grid()
    plt.show()


if __name__ == '__main__':
    run_experiments(timeout=2, partitions_n=5, n_jobs=-1)



from typing import Union, Iterable, Callable

from fedot.core.optimisers.objective import Objective
from fedot.core.repository.quality_metrics_repository import \
    MetricType, MetricsEnum, MetricsRepository, ComplexityMetricsEnum
from fedot.core.utilities.data_structures import ensure_wrapped_in_sequence


class MetricsObjective(Objective):
    def __init__(self,
                 metrics: Union[MetricType, Iterable[MetricType]],
                 is_multi_objective: bool = False):
        quality_metrics = {}
        complexity_metrics = {}

        for metric in ensure_wrapped_in_sequence(metrics):
            if isinstance(metric, MetricsEnum):
                metric_func = MetricsRepository().metric_by_id(metric)

                if isinstance(metric, ComplexityMetricsEnum):
                    complexity_metrics[metric] = metric_func
                else:
                    quality_metrics[metric] = metric_func
            elif isinstance(metric, Callable):
                metric_id = str(metric)
                quality_metrics[metric_id] = metric

        super().__init__(quality_metrics, complexity_metrics, is_multi_objective)



import logging
import os

import numpy as np
import pandas as pd

from fedot.api.main import Fedot
from fedot.core.repository.tasks import TsForecastingParams
from fedot.core.utils import fedot_project_root


def prepare_input_data(train_file_path, test_file_path, history_size: int = 10000):
    """ Function for preparing InputData for train and test algorithm

    :param train_file_path: path to the csv file for training
    :param test_file_path: path to the csv file for validation
    :param history_size: length of time series

    :return dataset_to_train: InputData for train
    :return dataset_to_validate: InputData for validation
    """
    # Load train and test dataframes
    full_path_train = os.path.join(str(fedot_project_root()), train_file_path)
    full_path_test = os.path.join(str(fedot_project_root()), test_file_path)
    df_train = pd.read_csv(full_path_train)
    df_test = pd.read_csv(full_path_test)

    ws_history = np.ravel(np.array(df_train['wind_speed']))[:history_size]
    ssh_history = np.ravel(np.array(df_train['sea_height']))[:history_size]
    ssh_obs = np.ravel(np.array(df_test['sea_height']))[:history_size]

    return ssh_history, ws_history, ssh_obs


def run_metocean_forecasting_problem(train_file_path, test_file_path,
                                     forecast_length=1, visualization=False, timeout=5):
    # Prepare data for train and test
    ssh_history, ws_history, ssh_obs = prepare_input_data(train_file_path, test_file_path)

    historical_data = {
        'ws': ws_history,  # additional variable
        'ssh': ssh_history,  # target variable
    }

    fedot = Fedot(problem='ts_forecasting',
                  task_params=TsForecastingParams(forecast_length=forecast_length),
                  timeout=timeout, logging_level=logging.DEBUG)

    pipeline = fedot.fit(features=historical_data, target=ssh_history)
    fedot.forecast(historical_data)
    metric = fedot.get_metrics(target=ssh_obs)

    if visualization:
        pipeline.show()
        fedot.plot_prediction(target='ssh')

    return metric


if __name__ == '__main__':
    # the dataset was obtained from NEMO model simulation for sea surface height

    # a dataset that will be used as a train and test set during composition
    file_path_train = 'cases/data/metocean/metocean_data_train.csv'

    # a dataset for a final validation of the composed model
    file_path_test = 'cases/data/metocean/metocean_data_test.csv'

    run_metocean_forecasting_problem(file_path_train, file_path_test,
                                     forecast_length=12, timeout=10,
                                     visualization=True)



import logging

import pandas as pd

from fedot.api.main import Fedot
from fedot.core.data.data import InputData
from fedot.core.data.data_split import train_test_data_setup
from fedot.core.pipelines.node import PrimaryNode, SecondaryNode
from fedot.core.pipelines.pipeline import Pipeline
from fedot.core.repository.dataset_types import DataTypesEnum
from fedot.core.repository.tasks import TsForecastingParams, Task, TaskTypesEnum
from fedot.core.utils import fedot_project_root

logging.raiseExceptions = False

datasets = {
    'australia': f'{fedot_project_root()}/examples/data/ts/australia.csv',
    'beer': f'{fedot_project_root()}/examples/data/ts/beer.csv',
    'salaries': f'{fedot_project_root()}/examples/data/ts/salaries.csv',
    'stackoverflow': f'{fedot_project_root()}/examples/data/ts/stackoverflow.csv'}


def run_ts_forecasting_example(dataset='australia', horizon: int = 30, validation_blocks=2, timeout: float = None,
                               visualization=False):
    time_series = pd.read_csv(datasets[dataset])

    task = Task(TaskTypesEnum.ts_forecasting,
                TsForecastingParams(forecast_length=horizon))
    if dataset not in ['australia']:
        idx = pd.to_datetime(time_series['idx'].values)
    else:
        # non datetime indexes
        idx = time_series['idx'].values
    time_series = time_series['value'].values
    train_input = InputData(idx=idx,
                            features=time_series,
                            target=time_series,
                            task=task,
                            data_type=DataTypesEnum.ts)
    train_data, test_data = train_test_data_setup(train_input, validation_blocks=validation_blocks)

    # init model for the time series forecasting
    model = Fedot(problem='ts_forecasting',
                  task_params=task.task_params,
                  timeout=timeout,
                  n_jobs=1,
                  cv_folds=2, validation_blocks=validation_blocks, preset='fast_train')

    # run AutoML model design in the same way
    pipeline = model.fit(train_data)

    # use model to obtain two-step in-sample forecast
    in_sample_forecast = model.predict(test_data)
    print('Metrics for two-step in-sample forecast: ',
          model.get_metrics(metric_names=['rmse', 'mae', 'mape']))

    # plot forecasting result
    if visualization:
        pipeline.show()
        model.plot_prediction()

    # use model to obtain one-step forecast
    train_data, test_data = train_test_data_setup(train_input)
    simple_forecast = model.forecast(test_data)
    print('Metrics for one-step forecast: ',
          model.get_metrics(metric_names=['rmse', 'mae', 'mape']))
    if visualization:
        model.plot_prediction()

    # use model to obtain two-step out-of-sample forecast
    out_of_sample_forecast = model.forecast(test_data, horizon=20)
    # we can not calculate metrics because we do not have enough future values
    if visualization:
        model.plot_prediction()

    return in_sample_forecast, simple_forecast, out_of_sample_forecast


if __name__ == '__main__':
    run_ts_forecasting_example(dataset='beer', horizon=10, timeout=2., visualization=True)



from typing import TYPE_CHECKING, Optional, Union

from fedot.core.caching.base_cache import BaseCache
from fedot.core.caching.preprocessing_cache_db import PreprocessingCacheDB
from fedot.utilities.debug import is_test_session

if TYPE_CHECKING:
    from fedot.core.pipelines.pipeline import Pipeline


class PreprocessingCache(BaseCache):
    """
    Usage of that class makes sense only if you have table data.
    Stores/loads `DataPreprocessor`'s encoders and imputers for pipelines to decrease optional preprocessing time.

    :param cache_folder: path to the place where cache files should be stored.
    """

    def __init__(self, cache_folder: Optional[str] = None):
        super().__init__(PreprocessingCacheDB(cache_folder))

    def try_load_preprocessor(self, pipeline: 'Pipeline', fold_id: Union[int, None]):
        """
        Tries to find preprocessor in DB table and load it for pipeline

        :param pipeline: pipeline to load preprocessor for
        :param fold_id: number of fold
        """
        try:
            structural_id = _get_db_uid(pipeline, fold_id)
            processors = self._db.get_preprocessor(structural_id)
            if processors:
                pipeline.encoder, pipeline.imputer = processors
        except Exception as ex:
            self.log.warning(f'Preprocessor search error: {ex}')
            if is_test_session():
                raise ex

    def add_preprocessor(self, pipeline: 'Pipeline', fold_id: Optional[Union[int, None]] = None):
        """
        Adds preprocessor into DB working table.

        :param pipeline: pipeline with preprocessor to add
        :param fold_id: number of fold
        """
        structural_id = _get_db_uid(pipeline, fold_id)
        self._db.add_preprocessor(structural_id, pipeline.preprocessor)


def _get_db_uid(pipeline: 'Pipeline', fold_id: Union[int, None]) -> str:
    """
    Constructs unique id from pipeline and data, which is considered as primary key for DB.

    :param pipeline: pipeline to get uid from
    :param fold_id: number of fold

    :return: unique pipeline plus related data identificator
    """
    fold_id = fold_id if fold_id is not None else ""
    pipeline_id = pipeline.root_node.descriptive_id
    return f'{pipeline_id}_{fold_id}'



import itertools
from typing import List, Optional, Dict

import numpy as np

from fedot.api.api_utils.presets import OperationsPreset
from fedot.core.repository.graph_operation_repository import GraphOperationRepository
from fedot.core.repository.operation_types_repository import get_operations_for_task
from fedot.core.repository.tasks import Task, TaskTypesEnum


class PipelineOperationRepository(GraphOperationRepository):
    """Repository in order to extract suitable operations by keys
     for pipelines during graph composition. Defines 2 keys:
     `primary` and `secondary` for distinguishing operations suitable
     for, respectively, primary and secondary nodes.

     Designed to work in cooperation with
     :py:class:`fedot.core.optimisers.opt_node_factory.OptNodeFactory`

     Args:
        operations_by_keys: available operations already splitted by keys
     """
    def __init__(self, operations_by_keys: Optional[Dict[str, List[str]]] = None):
        super().__init__()
        self.operations_by_keys = operations_by_keys or dict()

    def from_available_operations(self, task: Task, preset: str,
                                  available_operations: List[str]):
        """ Initialize repository from available operations, task and preset """
        operations_by_task_preset = OperationsPreset(task, preset).filter_operations_by_preset()
        all_operations = list(set.intersection(set(operations_by_task_preset), set(available_operations)))
        primary_operations, secondary_operations = \
            self.divide_operations(all_operations, task)
        self.operations_by_keys = {'primary': primary_operations, 'secondary': secondary_operations}

    def get_operations(self, is_primary: bool) -> List[str]:
        """ Get pipeline operations by specified model key """
        operation_key = 'primary' if is_primary else 'secondary'
        operations = self.operations_by_keys.get(operation_key, [])
        return operations

    def get_all_operations(self) -> List[str]:
        """ Get all pipeline operations with all keys """
        return list(itertools.chain(*self.operations_by_keys.values()))

    @staticmethod
    def divide_operations(available_operations, task):
        """ Function divide operations for primary and secondary """

        if task.task_type is TaskTypesEnum.ts_forecasting:
            # Get time series operations for primary nodes
            ts_data_operations = get_operations_for_task(task=task,
                                                         mode='data_operation',
                                                         tags=["non_lagged"])
            # Remove exog data operation from the list
            ts_data_operations.remove('exog_ts')

            ts_primary_models = get_operations_for_task(task=task,
                                                        mode='model',
                                                        tags=["non_lagged"])
            # Union of the models and data operations
            ts_primary_operations = ts_data_operations + ts_primary_models

            # Filter - remain only operations, which were in available ones
            primary_operations = list(set(ts_primary_operations).intersection(available_operations))
            secondary_operations = available_operations
        else:
            primary_operations = available_operations
            secondary_operations = available_operations
        return primary_operations, secondary_operations



from copy import deepcopy
from random import randint
from typing import Any, List, Optional, Tuple

from fedot.core.constants import MAXIMAL_ATTEMPTS_NUMBER
from fedot.core.dag.graph_utils import distance_to_root_level, distance_to_primary_level
from fedot.core.optimisers.gp_comp.pipeline_composer_requirements import PipelineComposerRequirements
from fedot.core.optimisers.graph import OptGraph, OptNode
from fedot.core.optimisers.opt_node_factory import OptNodeFactory
from fedot.core.optimisers.optimizer import GraphGenerationParams


def random_graph(graph_generation_params: GraphGenerationParams,
                 requirements: PipelineComposerRequirements,
                 max_depth: Optional[int] = None) -> OptGraph:
    max_depth = max_depth if max_depth else requirements.max_depth
    is_correct_graph = False
    graph = None
    n_iter = 0
    requirements = adjust_requirements(requirements)
    node_factory = graph_generation_params.node_factory

    while not is_correct_graph:
        graph = OptGraph()
        if requirements.max_depth == 1:
            graph_root = node_factory.get_node(is_primary=True)
            graph.add_node(graph_root)
        else:
            graph_root = node_factory.get_node(is_primary=False)
            graph.add_node(graph_root)
            graph_growth(graph, graph_root, node_factory, requirements, max_depth)

        is_correct_graph = graph_generation_params.verifier(graph)
        n_iter += 1
        if n_iter > MAXIMAL_ATTEMPTS_NUMBER:
            raise ValueError(f'Could not generate random graph for {n_iter} '
                             f'iterations with requirements {requirements}')
    return graph


def adjust_requirements(requirements: PipelineComposerRequirements) -> PipelineComposerRequirements:
    """Function returns modified copy of the requirements if necessary.
    Example: Graph with only one primary node should consist of only one primary node
    without duplication, because this causes errors. Therefore minimum and maximum arity
    become equal to one.
    """
    requirements = deepcopy(requirements)
    if len(requirements.primary) == 1 and requirements.max_arity > 1:
        requirements.min_arity = requirements.max_arity = 1
    return requirements


def graph_growth(graph: OptGraph,
                 node_parent: OptNode,
                 node_factory: OptNodeFactory,
                 requirements: PipelineComposerRequirements,
                 max_depth: int):
    """Function create a graph and links between nodes"""
    offspring_size = randint(requirements.min_arity, requirements.max_arity)

    for offspring_node in range(offspring_size):
        height = distance_to_root_level(graph, node_parent)
        is_max_depth_exceeded = height >= max_depth - 2
        is_primary_node_selected = height < max_depth - 1 and randint(0, 1)
        if is_max_depth_exceeded or is_primary_node_selected:
            primary_node = node_factory.get_node(is_primary=True)
            node_parent.nodes_from.append(primary_node)
            graph.add_node(primary_node)
        else:
            secondary_node = node_factory.get_node(is_primary=False)
            graph.add_node(secondary_node)
            node_parent.nodes_from.append(secondary_node)
            graph_growth(graph, secondary_node, node_factory, requirements, max_depth)


def equivalent_subtree(graph_first: Any, graph_second: Any) -> List[Tuple[Any, Any]]:
    """Finds the similar subtree in two given trees"""

    def structural_equivalent_nodes(node_first, node_second):
        nodes = []
        is_same_type = type(node_first) == type(node_second)
        node_first_childs = node_first.nodes_from
        node_second_childs = node_second.nodes_from
        if is_same_type and (not node_first.nodes_from or
                             node_first_childs and node_second_childs and
                             len(node_first_childs) == len(node_second_childs)):
            nodes.append((node_first, node_second))
            if node_first.nodes_from:
                for node1_child, node2_child in zip(node_first.nodes_from, node_second.nodes_from):
                    nodes_set = structural_equivalent_nodes(node1_child, node2_child)
                    if nodes_set:
                        nodes += nodes_set
        return nodes

    pairs_set = structural_equivalent_nodes(graph_first.root_node, graph_second.root_node)
    return pairs_set


def replace_subtrees(graph_first: Any, graph_second: Any, node_from_first: Any, node_from_second: Any,
                     layer_in_first: int, layer_in_second: int, max_depth: int):
    node_from_graph_first_copy = deepcopy(node_from_first)

    summary_depth = layer_in_first + distance_to_primary_level(node_from_second) + 1
    if summary_depth <= max_depth and summary_depth != 0:
        graph_first.update_subtree(node_from_first, node_from_second)

    summary_depth = layer_in_second + distance_to_primary_level(node_from_first) + 1
    if summary_depth <= max_depth and summary_depth != 0:
        graph_second.update_subtree(node_from_second, node_from_graph_first_copy)


def num_of_parents_in_crossover(num_of_final_inds: int) -> int:
    return num_of_final_inds if not num_of_final_inds % 2 else num_of_final_inds + 1


def filter_duplicates(archive, population) -> List[Any]:
    filtered_archive = []
    for ind in archive.items:
        has_duplicate_in_pop = False
        for pop_ind in population:
            if ind.fitness == pop_ind.fitness:
                has_duplicate_in_pop = True
                break
        if not has_duplicate_in_pop:
            filtered_archive.append(ind)
    return filtered_archive



from typing import List

from fedot.core.data.data import InputData, OutputData
from fedot.core.repository.tasks import TaskTypesEnum
from fedot.core.repository.quality_metrics_repository import (
    MetricsEnum,
    RegressionMetricsEnum,
    ClassificationMetricsEnum,
    ClusteringMetricsEnum,
    MetricsRepository
)


class MetricByTask:
    __metric_by_task = {TaskTypesEnum.regression: RegressionMetricsEnum.RMSE,
                        TaskTypesEnum.classification: ClassificationMetricsEnum.ROCAUC_penalty,
                        TaskTypesEnum.clustering: ClusteringMetricsEnum.silhouette,
                        TaskTypesEnum.ts_forecasting: RegressionMetricsEnum.RMSE,
                        }

    @staticmethod
    def get_default_quality_metrics(task_type: TaskTypesEnum) -> List[MetricsEnum]:
        return [MetricByTask.__metric_by_task.get(task_type)]

    @staticmethod
    def compute_default_metric(task_type: TaskTypesEnum, true: InputData, predicted: OutputData, round_up_to: int = 6) -> float:
        """Returns the value of metric defined by task"""
        metric_id = MetricByTask.get_default_quality_metrics(task_type)[0]
        metric = MetricsRepository.metric_class_by_id(metric_id)
        try:
            return round(metric.metric(reference=true, predicted=predicted), round_up_to)
        except ValueError:
            return metric.default_value
